import time
from unittest import TestCase, TestLoader, TextTestRunner

import cobra
from cobra import Model
from cobra.flux_analysis.gapfilling import GapFiller

from boimmgpy.database.accessors.compounds_database_accessor import CompoundsDBAccessor
from boimmgpy.boimmg.RepresentationRevisors.RepresentationSimpleCaseSolver import SimpleCaseSolver
from boimmgpy.boimmg.RepresentationRevisors.RepresentationRedundantCaseSolver import RepresentationRedundantCaseSolver
from boimmgpy.case_studies.redundant_representation_case.map_ecoli_model import map_iJR904


class TestQuinones(TestCase):


    def setUp(self):
        self.startTime = time.time()

    def tearDown(self):
        t = time.time() - self.startTime
        print('%s: %.9f' % (self.id(), t))

    def test_simulate_ecoli_ubiquinone10(self):

        model = cobra.io.read_sbml_model("C:/Users/Biosystems/Desktop/thesis_jcapela/BOIMMGpy/models/iML1515.xml")
        gapfiller = SimpleCaseSolver(model, "BiGG")
        gapfiller.swap_and_gap_fill("cpd08232")
        cobra.io.write_sbml_model(gapfiller.model,"iML1515_ubq10.xml")

    def test_simulate_ecoli_generalize(self):
        model = cobra.io.read_sbml_model("C:/Users/Biosystems/Desktop/thesis_jcapela/BOIMMGpy/models/iML1515.xml")
        gapfiller = SimpleCaseSolver(model, "BiGG")
        gapfiller.swap_compound("cpd15560","cpd11669")
        cobra.io.write_sbml_model(gapfiller.model, "simple_representation_case/iML1515_generalized.xml")

    def test_database_access(self):

        accessor = CompoundsDBAccessor()
        accessor.get_successors_by_ont_id_rel_type(3,"is_a")

class TestLipids(TestCase):


    def setUp(self):
        self.startTime = time.time()

    def tearDown(self):
        t = time.time() - self.startTime
        print('%s: %.9f' % (self.id(), t))

    def get_targets_to_replace_by_type(self, target: int):
        compounds_ontology = CompoundsDBAccessor()

        res={}
        precursors = compounds_ontology.get_all_predecessors_by_ont_id_rel_type(target,"precursor_of")

        for precursor in precursors:
            parent = compounds_ontology.get_successors_by_ont_id_rel_type(precursor,"is_a")

            if parent[0] in res:
                res.get(parent[0]).append(precursor)
            else:
                res[parent[0]] = [precursor]
        return res

    def test_precursors(self):
        compounds_ontology = CompoundsDBAccessor()
        all_precursors = compounds_ontology.get_all_predecessors_by_ont_id_rel_type(322790,"precursor_of")
        print(all_precursors)


    def test_simulate_ecoli_ubiquinone10(self):

        model = cobra.io.read_sbml_model("C:/Users/Biosystems/Desktop/thesis_jcapela/BOIMMGpy/models/iML1515.xml")
        gapfiller = SimpleCaseSolver(model, "BiGG")
        gapfiller.swap_and_gap_fill(690028)

    def test_components_scraping(self):
        db_acessor = CompoundsDBAccessor()
        components = [200,437,436,129]

        all_new_met = []
        compounds = db_acessor.get_compounds_with_specific_parent_within_set_of_components(12, components)
        for compound in compounds:
            precursors = db_acessor.get_all_predecessors_by_ont_id_rel_type(compound,"precursor_of")
            all_new_met.extend(precursors)

        # res = []
        # for new_met in all_new_met:
        #     if new_met not in res:
        #         res.append(new_met)

        print(len(all_new_met))
        # print(res)

    def test_identification_of_biosynthesis_pathway_ecoli(self):
        model = map_iJR904()

        solver = RepresentationRedundantCaseSolver(model,"BiGG")

        reactions,other = solver.identify_biosynthesis_pathway(12)

        assert len(reactions) == 10
        # solver.swap_from_generic(12,[200,437,436,129])

    def test_dict_with_predecessors(self):
        model = map_iJR904()

        solver = RepresentationRedundantCaseSolver(model, "BiGG")
        res = solver.get_targets_to_replace_by_type(300374)

        print()


    def test_phospholipid_granulation_ecoli(self):

        model = map_iJR904()

        components = ["cpd00214", "cpd01080", "cpd03847", "cpd05274"]

        solver = RepresentationRedundantCaseSolver(model, "BiGG")

        solver.swap_from_generic(["cpd22513","cpd15649"], components, True)

        cobra.io.write_sbml_model(solver.model, "../models/enhanced_model_ecoli.xml")
        cobra.io.save_json_model(model, "enhanced_model_ecoli.json")

    def test_components_reactions_solver(self):
        model = map_iJR904()

        solver = RepresentationRedundantCaseSolver(model, "BiGG")

        solver.solve_components_reactions(True)


    def test_optimization(self):

        model = cobra.io.read_sbml_model("../models/enhanced_model_ecoli.xml")

        gapfiller = GapFiller(model, Model(),
                              demand_reactions=True,
                              exchange_reactions=True,
                              integer_threshold=1e-300)

        solution = gapfiller.fill(iterations=1)
        print(solution)
        model.add_reactions(solution[0])

        print(model.optimize().objective_value)

        cobra.io.write_sbml_model(model,"enhanced_model_ecoli2.xml")
        cobra.io.save_json_model(model,"enhanced_model_ecoli.json")

    def test_read_universal_model(self):

        model = cobra.io.read_sbml_model("C:/Users/BioSystems/Desktop/universal_model.xml")

    def test_algae_model(self):
        model = cobra.io.read_sbml_model("../models/cvulgaris.xml")

        model.objective = model.reactions.get_by_id("e_Biomass__cytop")

        r_acyl = model.reactions.get_by_id("e_Fatty_acid__cytop")
        reactants = r_acyl.reactants

        components_list = []
        for reactant in reactants:

            if "kegg.compound" in reactant.annotation:
                compound = reactant.annotation.get("kegg.compound")
                components_list.append(compound)


        ### confirmar se há alguma reação com fatty acid

        solver = RepresentationRedundantCaseSolver(model, "KEGG")

        # components = ["lgnc","arach","C06427","clpnd"]

        solver.swap_from_generic(["C00422","C06037"],components_list,True)

        cobra.io.write_sbml_model(model, "very_nice_algae_model.xml")

if __name__ == '__main__':
    suite = TestLoader().loadTestsFromTestCase()
    TextTestRunner(verbosity=0).run(suite)


import cobra
from cobra import Reaction, Model
from cobra.flux_analysis.gapfilling import GapFiller

from boimmgpy.service.representation_problem_solvers.representation_redundant_case_solver import RedundantCaseSolver

def redundant_granulator_yeast():
    model = cobra.io.read_sbml_model("mapped_yeast.xml")
    components = ["C08362","C00712","C00249","C01530"]

    solver = RedundantCaseSolver(model, "KEGG")
    solver.map_model()
    # solver.load_maps("../../dumps/")

    solver.swap_from_generic(["cpd22513", "C00422"], components, False)
    solver.generateISAreactions()

    solver.write_reports("new_report.txt")
    cobra.io.write_sbml_model(solver.model, "granulated_yeast.xml")




def redundant_granulator_ecoli():
    # model = map_iJR904()
    model = cobra.io.read_sbml_model("iJR904_mapped.xml")
    components = ["cpd00214", "cpd03847", "cpd05274","cpd25615","cpd05237"]

    solver = RedundantCaseSolver(model, "BiGG")
    solver.map_model()

    solver.swap_from_generic(["cpd22513", "cpd15649"], components, True)
    solver.generateISAreactions()

    solver.write_reports("new_report.txt")
    cobra.io.write_sbml_model(solver.model, "granulated_iJR904.xml")

    PASYN_EC(model)
    LPLIPA2(model)
    LPLIPA1(model)
    LPLIPA3(model)
    PLIPA1(model)

    cobra.io.write_sbml_model(solver.model, "granulated_gap_filled_iJR904.xml")

def new_test_server():
    model = cobra.io.read_sbml_model("new_model.xml")
    PASYN_EC(model)
    LPLIPA2(model)
    LPLIPA1(model)
    LPLIPA3(model)
    PLIPA1(model)
    cobra.io.write_sbml_model(model, "new_model2.xml")


def PLIPA3(model):
    s1 = {model.metabolites.get_by_id("BMGC427_c"): -1,
          model.metabolites.get_by_id("hdca_c"): 1,
          model.metabolites.get_by_id("BMGC314_c"): 1}

    s2 = {model.metabolites.get_by_id("BMGC7454_c"): -1,
          model.metabolites.get_by_id("hdcea_c"): 1,
          model.metabolites.get_by_id("BMGC16363_c"): 1}

    s3 = {model.metabolites.get_by_id("BMGC7408_c"): -1,
          model.metabolites.get_by_id("ocdcea_c"): 1,
          model.metabolites.get_by_id("BMGC16317_c"): 1}

    s4 = {model.metabolites.get_by_id("BMGC7474_c"): -1,
          model.metabolites.get_by_id("ttdca_c"): 1,
          model.metabolites.get_by_id("BMGC452_c"): 1}

    s5 = {model.metabolites.get_by_id("BMGC7456_c"): -1,
          model.metabolites.get_by_id("ttdcea_c"): 1,
          model.metabolites.get_by_id("BMGC16364_c"): 1}

    r1 = Reaction("PLIPA3_1")
    r2 = Reaction("PLIPA3_2")
    r3 = Reaction("PLIPA3_3")
    r4 = Reaction("PLIPA3_4")
    r5 = Reaction("PLIPA3_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("h2o_c")] = -1
        si[model.metabolites.get_by_id("h_c")] = 1

        ri.add_metabolites(si)

    model.add_reactions(r)


def PLIPA1(model):
    s1 = {model.metabolites.get_by_id("BMGC427_c"): -1,
          model.metabolites.get_by_id("hdca_c"): 1,
          model.metabolites.get_by_id("BMGC314_c"): 1}

    s2 = {model.metabolites.get_by_id("BMGC7454_c"): -1,
          model.metabolites.get_by_id("hdcea_c"): 1,
          model.metabolites.get_by_id("BMGC16363_c"): 1}

    s3 = {model.metabolites.get_by_id("BMGC7408_c"): -1,
          model.metabolites.get_by_id("ocdcea_c"): 1,
          model.metabolites.get_by_id("BMGC16317_c"): 1}

    s4 = {model.metabolites.get_by_id("BMGC7474_c"): -1,
          model.metabolites.get_by_id("ttdca_c"): 1,
          model.metabolites.get_by_id("BMGC452_c"): 1}

    s5 = {model.metabolites.get_by_id("BMGC7456_c"): -1,
          model.metabolites.get_by_id("ttdcea_c"): 1,
          model.metabolites.get_by_id("BMGC16364_c"): 1}

    r1 = Reaction("PLIPA1_1")
    r2 = Reaction("PLIPA1_2")
    r3 = Reaction("PLIPA1_3")
    r4 = Reaction("PLIPA1_4")
    r5 = Reaction("PLIPA1_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("h2o_c")] = -1
        si[model.metabolites.get_by_id("h_c")] = 1

        ri.add_metabolites(si)

    model.add_reactions(r)

def LPLIPA1(model):
    s1 = {model.metabolites.get_by_id("BMGC314_c"): -1,
          model.metabolites.get_by_id("hdca_c"): 1}

    s2 = {model.metabolites.get_by_id("BMGC16363_c"): -1,
          model.metabolites.get_by_id("hdcea_c"): 1}

    s3 = {model.metabolites.get_by_id("BMGC16317_c"): -1,
          model.metabolites.get_by_id("ocdcea_c"): 1}

    s4 = {model.metabolites.get_by_id("BMGC452_c"): -1,
          model.metabolites.get_by_id("ttdca_c"): 1}

    s5 = {model.metabolites.get_by_id("BMGC16364_c"): -1,
          model.metabolites.get_by_id("ttdcea_c"): 1}

    r1 = Reaction("LPLIPA1_1")
    r2 = Reaction("LPLIPA1_2")
    r3 = Reaction("LPLIPA1_3")
    r4 = Reaction("LPLIPA1_4")
    r5 = Reaction("LPLIPA1_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("h2o_c")] = -1
        si[model.metabolites.get_by_id("h_c")] = 1
        si[model.metabolites.get_by_id("g3pg_c")] = 1

        ri.add_metabolites(si)

    model.add_reactions(r)

def LPLIPA3(model):
    s1 = {model.metabolites.get_by_id("BMGC223_c"): -1,
          model.metabolites.get_by_id("hdca_c"): 1}

    s2 = {model.metabolites.get_by_id("BMGC6942_c"): -1,
          model.metabolites.get_by_id("hdcea_c"): 1}

    s3 = {model.metabolites.get_by_id("BMGC6896_c"): -1,
          model.metabolites.get_by_id("ocdcea_c"): 1}

    s4 = {model.metabolites.get_by_id("BMGC319_c"): -1,
          model.metabolites.get_by_id("ttdca_c"): 1}

    s5 = {model.metabolites.get_by_id("BMGC6943_c"): -1,
          model.metabolites.get_by_id("ttdcea_c"): 1}

    r1 = Reaction("LPLIPA3_1")
    r2 = Reaction("LPLIPA3_2")
    r3 = Reaction("LPLIPA3_3")
    r4 = Reaction("LPLIPA3_4")
    r5 = Reaction("LPLIPA3_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("h2o_c")] = -1
        si[model.metabolites.get_by_id("h_c")] = 1
        si[model.metabolites.get_by_id("g3pc_c")] = 1

        ri.add_metabolites(si)

    model.add_reactions(r)

def LPLIPA2(model):
    s1 = {model.metabolites.get_by_id("BMGC291_c"): -1,
          model.metabolites.get_by_id("hdca_c"): 1}

    s2 = {model.metabolites.get_by_id("BMGC27923_c"): -1,
          model.metabolites.get_by_id("hdcea_c"): 1}

    s3 = {model.metabolites.get_by_id("BMGC27877_c"): -1,
          model.metabolites.get_by_id("ocdcea_c"): 1}

    s4 = {model.metabolites.get_by_id("BMGC27943_c"): -1,
          model.metabolites.get_by_id("ttdca_c"): 1}

    s5 = {model.metabolites.get_by_id("BMGC27925_c"): -1,
          model.metabolites.get_by_id("ttdcea_c"): 1}

    r1 = Reaction("LPLIPA2_1")
    r2 = Reaction("LPLIPA2_2")
    r3 = Reaction("LPLIPA2_3")
    r4 = Reaction("LPLIPA2_4")
    r5 = Reaction("LPLIPA2_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("h2o_c")] = -1
        si[model.metabolites.get_by_id("h_c")] = 1
        si[model.metabolites.get_by_id("g3pe_c")] = 1

        ri.add_metabolites(si)

    model.add_reactions(r)

def PASYN_EC(model):
    s1 = {model.metabolites.get_by_id("palmACP_c"): -1,
          model.metabolites.get_by_id("BMGC423_c"): 1}

    s2 = {model.metabolites.get_by_id("myrsACP_c"): -1,
          model.metabolites.get_by_id("BMGC38435_c"): 1}

    s3 = {model.metabolites.get_by_id("hdeACP_c"): -1,
          model.metabolites.get_by_id("BMGC38416_c"): 1}

    s4 = {model.metabolites.get_by_id("octeACP_c"): -1,
          model.metabolites.get_by_id("BMGC38371_c"): 1}

    s5 = {model.metabolites.get_by_id("tdeACP_c"): -1,
          model.metabolites.get_by_id("BMGC38417_c"): 1}

    r1 = Reaction("reaction_1")
    r2 = Reaction("reaction_2")
    r3 = Reaction("reaction_3")
    r4 = Reaction("reaction_4")
    r5 = Reaction("reaction_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("glyc3p_c")] = -1
        si[model.metabolites.get_by_id("ACP_c")] = 2

        ri.add_metabolites(si)

    model.add_reactions(r)

if __name__ == "__main__":
    redundant_granulator_yeast()
    # redundant_granulator_ecoli()
    # model = cobra.io.read_sbml_model("granulated_gap_filled_iJR904.xml")
    #
    # reactions = []
    # reaction = model.reactions.get_by_id("PASYN_EC")
    # reactions.append(reaction)
    # reaction = model.reactions.get_by_id("LPLIPA2")
    # reactions.append(reaction)
    #
    # reaction = model.reactions.get_by_id("LPLIPA3")
    # reactions.append(reaction)
    #
    # reaction = model.reactions.get_by_id("LPLIPA1")
    # reactions.append(reaction)
    #
    # model.remove_reactions(reactions)

    # reaction = model.reactions.get_by_id("PLIPA1")
    # reactions.append(reaction)

    # cobra.io.save_json_model(model,"new_model_2.json")



    # model.solver = 'cplex'

    # gp = GapFiller(model, exchange_reactions=True, demand_reactions=True, integer_threshold=1e-1000)
    # solution = gp.fill(1)
    #
    # print(solution)
    #
    # model.add_reactions(solution[0])

    # print(model.optimize().objective_value)
    # cobra.io.write_sbml_model(model,"new_model2.xml")

    # new_test_server()


import json

from boimmgpy.database.accessors.compounds_database_accessor import CompoundsDBAccessor
from boimmgpy.definitions import COMPOUNDS_ANNOTATION_CONFIGS_PATH
from boimmgpy.utilities import file_utilities


class ModelMapper:

    def __init__(self,model,compoundsIdConverter,accessor):

        self.model = model
        self.__compoundsAnnotationConfigs = file_utilities.read_conf_file(
            COMPOUNDS_ANNOTATION_CONFIGS_PATH)
        self.__compoundsIdConverter = compoundsIdConverter
        self.__compounds_ontology = accessor

        self.mapped = False

        self.boimmg_db_model_map = {}

        self.boimmg_db_model_map_reverse = {}

        self.compounds_aliases_indexation = {}

        self.compound_inchikey_indexation = {}

        self.__compounds_aliases_indexation_reverse = {}

    @property
    def model(self):
        return self.__model

    @model.setter
    def model(self, value):
        self.__model = value

    @property
    def boimmg_db_model_map(self):
        return self.__boimmg_db_model_map

    @boimmg_db_model_map.setter
    def boimmg_db_model_map(self, value):
        if isinstance(value, dict):
            self.__boimmg_db_model_map = value
        else:
            raise ValueError("introduce a dictionary")

    @property
    def boimmg_db_model_map_reverse(self):
        return self.__boimmg_db_model_map_reverse

    @boimmg_db_model_map_reverse.setter
    def boimmg_db_model_map_reverse(self,value):
        if isinstance(value, dict):
            self.__boimmg_db_model_map_reverse = value
        else:
            raise ValueError("introduce a dictionary")

    @property
    def compounds_aliases_indexation(self):
        return self.__compounds_aliases_indexation

    @compounds_aliases_indexation.setter
    def compounds_aliases_indexation(self, value):
        if isinstance(value, dict):
            self.__compounds_aliases_indexation = value
        else:
            raise ValueError("introduce a dictionary")

    @property
    def compounds_aliases_indexation_reverse(self):
        return self.__compounds_aliases_indexation_reverse

    @compounds_aliases_indexation_reverse.setter
    def compounds_aliases_indexation_reverse(self, value):
        if isinstance(value, dict):
            self.__compounds_aliases_indexation_reverse = value
        else:
            raise ValueError("introduce a dictionary")

    @property
    def compound_inchikey_indexation(self):
        return self.__compound_inchikey_indexation

    @compound_inchikey_indexation.setter
    def compound_inchikey_indexation(self, value):
        if isinstance(value, dict):
            self.__compound_inchikey_indexation = value
        else:
            raise ValueError("introduce a dictionary")

    @property
    def mapped(self):
        return self._mapped

    @mapped.setter
    def mapped(self,value: bool):
        self._mapped = value

    def printProgressBar(self,iteration, total, prefix='', suffix='', decimals=1, length=100, fill='|', printEnd="\r"):
        """
      Call in a loop to create terminal progress bar
      @params:
          iteration   - Required  : current iteration (Int)
          total       - Required  : total iterations (Int)
          prefix      - Optional  : prefix string (Str)
          suffix      - Optional  : suffix string (Str)
          decimals    - Optional  : positive number of decimals in percent complete (Int)
          length      - Optional  : character length of bar (Int)
          fill        - Optional  : bar fill character (Str)
          printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
      """
        percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
        filledLength = int(length * iteration // total)
        bar = fill * filledLength + '-' * (length - filledLength)
        print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end="", flush=True)
        # Print New Line on Complete
        if iteration == total:
            print()

    def map_model(self,database):

        if database in self.__compoundsAnnotationConfigs.keys():
            database_annotation_key = self.__compoundsAnnotationConfigs[database]
        else:
            raise ValueError("introduce a valid database name")

        j=0
        model_metabolites = list(self.model.metabolites)
        for compound in self.model.metabolites:

            self.printProgressBar(j,len(model_metabolites))
            j+=1
            compound_annotation = compound.annotation

            # boimmg_compound_found = False

            annotation_keys = ["seed.compound",database_annotation_key,"boimmg.compound","inchi_key"]

            compound_id = compound.id

            compound_id = compound_id.split("_")[0]

            boimmg_compound_found = self.map_compound(compound_id,compound)

            for key in annotation_keys:
                if key in compound_annotation.keys():

                    if isinstance(compound_annotation[key], str):
                        annotation_values = [compound_annotation[key]]
                    else:
                        annotation_values = compound_annotation[key]

                    for annotation_value in annotation_values:

                        if not boimmg_compound_found:

                            boimmg_compound_found = self.map_compound(annotation_value,compound)

                        if key not in self.compounds_aliases_indexation:
                            self.compounds_aliases_indexation[key] = {}

                        if annotation_value not in self.compounds_aliases_indexation[key]:
                            self.compounds_aliases_indexation[key][annotation_value] = []

                        if compound.id not in self.compounds_aliases_indexation_reverse:
                            self.compounds_aliases_indexation_reverse[compound.id] = {}

                        if key not in self.compounds_aliases_indexation_reverse[compound.id]:
                            self.compounds_aliases_indexation_reverse[compound.id] = {key:[]}

                        if annotation_value not in self.compounds_aliases_indexation_reverse[compound.id][key]:
                            self.compounds_aliases_indexation_reverse[compound.id][key].append(annotation_value)

                        if compound.id not in self.compounds_aliases_indexation_reverse[compound.id][key]:
                            self.compounds_aliases_indexation[key][annotation_value].append(compound.id)

                    if not boimmg_compound_found:

                        if key == "boimmg.compound":
                            construction_sub_string = self.__compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]
                            value = compound_annotation[key]

                            boimmg_id = int(value.replace(construction_sub_string, ""))

                            if boimmg_id not in self.boimmg_db_model_map_reverse.keys():
                                self.boimmg_db_model_map_reverse[boimmg_id] = []

                            self.boimmg_db_model_map_reverse[boimmg_id].append(compound.id)

                            if boimmg_id not in self.boimmg_db_model_map.keys():
                                self.boimmg_db_model_map_reverse[compound.id] = []

                            self.boimmg_db_model_map_reverse[compound.id].append(compound.id)

                        elif key == "inchi_key":
                            value = compound_annotation[key][:-1]
                            if value not in self.compound_inchikey_indexation:
                                self.compound_inchikey_indexation[value] = [compound.id]
                            else:
                                self.compound_inchikey_indexation[value].append(compound.id)

        self.mapped = True

    def map_compound(self, annotation_value,compound):

        boimmg_compound_found = False

        if "cpd" in annotation_value:
            model_seed_id = [annotation_value]
        else:
            model_seed_id = self.__compoundsIdConverter.convert_db_id_to_model_seed_by_db_id(annotation_value)

        if model_seed_id:
            found = False
            i = 0
            while not found and i < len(model_seed_id):
                node = self.__compounds_ontology.get_node_id_from_model_seed_id(model_seed_id[i])
                i += 1
                if node:
                    found = True

                    boimmg_compound_found = True
                    self.boimmg_db_model_map[compound.id] = node

                    if node in self.boimmg_db_model_map_reverse.keys():
                        self.boimmg_db_model_map_reverse[node].append(compound.id)

                    else:
                        self.boimmg_db_model_map_reverse[node] = [compound.id]

        return boimmg_compound_found

    def create_map_dump(self,destination):

        out_file = open(destination + "boimmg_db_model_map.json", "w")
        json.dump(self.boimmg_db_model_map,out_file)
        out_file.close()

        out_file = open(destination + "boimmg_db_model_map_reverse.json", "w")
        json.dump(self.boimmg_db_model_map_reverse, out_file)
        out_file.close()

        out_file = open(destination + "compounds_aliases_indexation.json", "w")
        json.dump(self.compounds_aliases_indexation, out_file)
        out_file.close()

        out_file = open(destination + "compounds_aliases_indexation_reverse.json", "w")
        json.dump(self.compounds_aliases_indexation_reverse, out_file)
        out_file.close()

        out_file = open(destination + "compound_inchikey_indexation.json", "w")
        json.dump(self.compound_inchikey_indexation, out_file)
        out_file.close()


    def upload_maps(self,folder):

        in_file = open(folder + "boimmg_db_model_map.json", "r")
        self.boimmg_db_model_map = json.load(in_file)
        in_file.close()

        in_file = open(folder + "boimmg_db_model_map_reverse.json", "r")
        temp = json.load(in_file)
        in_file.close()
        self.boimmg_db_model_map_reverse = {int(k): v for k, v in temp.items()}

        in_file = open(folder + "compounds_aliases_indexation.json", "r")
        self.compounds_aliases_indexation = json.load(in_file)
        in_file.close()

        in_file = open(folder + "compounds_aliases_indexation_reverse.json", "r")
        self.compounds_aliases_indexation_reverse = json.load(in_file)
        in_file.close()

        in_file = open(folder + "compound_inchikey_indexation.json", "r")
        self.compound_inchikey_indexation = json.load(in_file)
        in_file.close()

        self.mapped = True

    def add_new_reactions_to_model(self,new_reactions):

        for reaction in new_reactions:

            self.add_new_metabolites_to_maps(reaction.metabolites)


    def __add_metabolites_to_boimmg_indexation(self,key,annotation_value,metabolite):

        if key == self.__compoundsAnnotationConfigs.get("ModelSEED"):
            model_seed_id = [annotation_value]

        else:
            model_seed_id = self.__compoundsIdConverter.convert_db_id_to_model_seed_by_db_id(
                annotation_value)

        if model_seed_id:

            found = False
            i = 0
            while not found and i < len(model_seed_id):
                node = self.__compounds_ontology.get_node_id_from_model_seed_id(model_seed_id[i])
                i += 1
                if node:
                    # found = True

                    self.boimmg_db_model_map[metabolite.id] = node

                    if node in self.boimmg_db_model_map_reverse.keys():
                        self.boimmg_db_model_map_reverse[node].append(metabolite.id)

                    else:
                        self.boimmg_db_model_map_reverse[node] = [metabolite.id]

                    return True

        return False

    def __add_metabolites_to_aliases_indexation(self,annotation_pair,annotation_value,metabolite):

        if annotation_pair not in self.compounds_aliases_indexation:
            self.compounds_aliases_indexation[annotation_pair] = {}

        if annotation_value not in self.compounds_aliases_indexation[annotation_pair]:
            self.compounds_aliases_indexation[annotation_pair][annotation_value] = []

        if metabolite.id not in self.compounds_aliases_indexation_reverse:
            self.compounds_aliases_indexation_reverse[metabolite.id] = {}

        if annotation_pair not in self.compounds_aliases_indexation_reverse[metabolite.id]:
            self.compounds_aliases_indexation_reverse[metabolite.id] = {annotation_pair: []}

        if annotation_value not in self.compounds_aliases_indexation_reverse[metabolite.id][annotation_pair]:
            self.compounds_aliases_indexation_reverse[metabolite.id][annotation_pair].append(annotation_value)

        if metabolite not in self.compounds_aliases_indexation[annotation_pair][annotation_value]:
            self.compounds_aliases_indexation[annotation_pair][annotation_value].append(metabolite.id)

    def add_new_metabolites_to_maps(self,new_metabolites):

        for metabolite in new_metabolites:

            compound_annotation = metabolite.annotation

            boimmg_compound_found = False
            for key in compound_annotation.keys():
                if key in self.__compoundsAnnotationConfigs.keys():
                    annotation_pair = self.__compoundsAnnotationConfigs[key]

                    if isinstance(compound_annotation[key], str):
                        annotation_values = [compound_annotation[key]]
                    else:
                        annotation_values = compound_annotation[key]

                    for annotation_value in annotation_values:

                        if not boimmg_compound_found:


                            boimmg_compound_found = self.__add_metabolites_to_boimmg_indexation(key,
                                                                                                annotation_value,
                                                                                                metabolite)

                        self.__add_metabolites_to_aliases_indexation(annotation_pair,annotation_value,metabolite)


                    if not boimmg_compound_found:

                        if key == "boimmg.compound":
                            construction_sub_string = self.__compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]
                            value = compound_annotation[key]

                            boimmg_id = int(value.replace(construction_sub_string, ""))

                            self.boimmg_db_model_map[metabolite.id] = boimmg_id

                            if boimmg_id not in self.boimmg_db_model_map_reverse.keys():
                                self.boimmg_db_model_map_reverse[boimmg_id] = [metabolite.id]

                            else:
                                self.boimmg_db_model_map_reverse[boimmg_id].append(metabolite.id)

                        elif key == "inchi_key":
                            value = compound_annotation[key][:-1]
                            if value not in self.compound_inchikey_indexation:
                                self.compound_inchikey_indexation[value] = [metabolite.id]
                            else:
                                self.compound_inchikey_indexation[value].append(metabolite.id)

    def update_maps(self, old_inchikey, new_inchikey, old_id, new_id, new_ontology_id, old_aliases, new_aliases):

        compound_container = self.model.metabolites.get_by_id(new_id)

        if old_id in self.boimmg_db_model_map:
            del self.boimmg_db_model_map[old_id]
            self.boimmg_db_model_map[new_id] = new_ontology_id

        for key,value in self.boimmg_db_model_map_reverse.items():
            if old_id in value:
                del self.boimmg_db_model_map_reverse[key]
                break

        if new_ontology_id not in self.boimmg_db_model_map_reverse:
            self.boimmg_db_model_map_reverse[new_ontology_id] = [compound_container.id]

        elif compound_container.id not in self.boimmg_db_model_map_reverse[new_ontology_id]:
            self.boimmg_db_model_map_reverse[new_ontology_id].append(compound_container.id)



        if old_inchikey:
            if old_inchikey[:-1] in self.compound_inchikey_indexation:
                del self.compound_inchikey_indexation[old_inchikey[:-1]]

        if new_inchikey:
            if new_inchikey[:-1] in self.compound_inchikey_indexation:
                self.compound_inchikey_indexation[new_inchikey[:-1]].append(compound_container.id)
            else:
                self.compound_inchikey_indexation[new_inchikey[:-1]] = [compound_container.id]

        self.__update_aliases_indexation(old_id,new_aliases,compound_container)


    def __update_aliases_indexation(self,old_id,new_aliases,compound_container):

        if old_id in self.compounds_aliases_indexation_reverse:
            old_aliases = self.compounds_aliases_indexation_reverse[old_id]
            for db in self.compounds_aliases_indexation:
                if db in old_aliases:
                    for alias in old_aliases[db]:
                        if alias in self.compounds_aliases_indexation[db]:
                            del self.compounds_aliases_indexation[db][alias]

            del self.compounds_aliases_indexation_reverse[old_id]

        for key in new_aliases:
            if key in self.__compoundsAnnotationConfigs.keys() and \
                    key in self.compounds_aliases_indexation:

                for alias in new_aliases[key]:
                    if alias in self.compounds_aliases_indexation[key]:
                        self.compounds_aliases_indexation[key][alias].append(compound_container.id)
                    else:
                        self.compounds_aliases_indexation[key][alias] = [compound_container.id]

    def check_metabolites_in_model(self, inchikey: str, aliases: dict, boimmg_container=None, boimmg_id = None) -> list:
        """
        This method checks whether a given metabolite with the :param inchikey and the :param aliases.
        If a BOIMMG node is available the search starts with it.

        :param str inchikey: InchiKey of the metabolite to be searched.
        :param dict aliases: databases links of the metabolite to be searched.
        :param CompoundNode boimmg_container: BOIMMG node
        :return list: metabolites in model
        """

        if not boimmg_container and not boimmg_id:

            if inchikey:
                inchikey_without_protonation = inchikey[:-1]
                if inchikey_without_protonation in self.compound_inchikey_indexation.keys():
                    compound_ids = self.compound_inchikey_indexation[inchikey_without_protonation]
                    compounds_in_model = [self.model.metabolites.get_by_id(compound_id)
                                          for compound_id in compound_ids]

                    return compounds_in_model

            for key in aliases:
                new_key = None
                if key in self.__compoundsAnnotationConfigs:
                    new_key = self.__compoundsAnnotationConfigs[key]

                if key in self.compounds_aliases_indexation:
                    for alias in aliases[key]:
                        if alias in self.compounds_aliases_indexation.get(key):

                            compound_ids = self.compounds_aliases_indexation[key].get(alias)
                            compounds_in_model = [self.model.metabolites.get_by_id(compound_id)
                                                  for compound_id in compound_ids]

                            return compounds_in_model

                elif new_key and new_key in self.compounds_aliases_indexation:
                    for alias in aliases[key]:
                        if alias in self.compounds_aliases_indexation.get(new_key):

                            compound_ids = self.compounds_aliases_indexation[new_key].get(alias)
                            compounds_in_model = [self.model.metabolites.get_by_id(compound_id)
                                                  for compound_id in compound_ids]

                            return compounds_in_model

        elif boimmg_container:
            ont_id = boimmg_container.id
            res = []

            if ont_id in self.__boimmg_db_model_map_reverse.keys():
                for model_id in self.__boimmg_db_model_map_reverse[ont_id]:
                    res.append(self.model.metabolites.get_by_id(model_id))

            return res

        else:
            res = []

            if boimmg_id in self.__boimmg_db_model_map_reverse.keys():
                for model_id in self.__boimmg_db_model_map_reverse[boimmg_id]:
                    res.append(self.model.metabolites.get_by_id(model_id))

            return res

        return []

    def check_if_boimmg_metabolite_in_model(self, boimmg_id, aliases = {} ):

        if boimmg_id in self.boimmg_db_model_map_reverse.keys():
            return self.boimmg_db_model_map_reverse[boimmg_id]

        elif aliases:
            for db in aliases:
                db_aliases = aliases[db]

                for alias in db_aliases:
                    if db in self.compounds_aliases_indexation.keys():
                        if alias in self.compounds_aliases_indexation[db]:
                            compounds = self.compounds_aliases_indexation[db][alias]
                            return compounds
        else:
            return None

        return None

    def get_boimmg_id_from_model_compound_id(self,model_id):

        if model_id in self.boimmg_db_model_map.keys():
            return self.boimmg_db_model_map[model_id]

        return None




import cobra

from boimmgpy import definitions
from boimmgpy.database.accessors.compounds_database_accessor import set_database_information
from boimmgpy.database.accessors.compounds_rest_accessor import CompoundsRestAccessor
from boimmgpy.definitions import ROOT_DIR
from boimmgpy.service.representation_problem_solvers.representation_simple_case_solver import SimpleCaseSolver


def sacc_to_9():
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iMM904.xml")
    set_database_information(uri="bolt://palsson.di.uminho.pt:6094",user="neo4j",password="Tuning999")

    rep_case_solver = SimpleCaseSolver(model, "ModelSEED")
    rep_case_solver.map_model()
    # rep_case_solver.dump_maps(ROOT_DIR + "/dumps/")
    # rep_case_solver.load_maps(ROOT_DIR + "/dumps/")
    rep_case_solver.swap_compound("cpd15290", "679986")
    rep_case_solver.write_report("test_report.txt")
    cobra.io.write_sbml_model(rep_case_solver.model, "iMM904_9.xml")

def sacc_to_rest():
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iMM904.xml")
    #set_database_information(uri="bolt://palsson.di.uminho.pt:6094",user="neo4j",password="Tuning999")

    rest_accessor = CompoundsRestAccessor()
    rep_case_solver = SimpleCaseSolver(model, "ModelSEED",rest_accessor)
    rep_case_solver.map_model()
    # rep_case_solver.dump_maps(ROOT_DIR + "/dumps/")
    # rep_case_solver.load_maps(ROOT_DIR + "/dumps/")
    rep_case_solver.swap_compound("cpd15290", "679986")
    rep_case_solver.write_report("test_report.txt")
    cobra.io.write_sbml_model(rep_case_solver.model, "iMM904_9.xml")

def simulate_ecoli_generalize():
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iML1515.xml")
    rep_case_solver = SimpleCaseSolver(model, "BiGG")
    rep_case_solver.swap_compound("cpd15560", "cpd11669")
    rep_case_solver.swap_compound("cpd15500", "cpd11451")
    cobra.io.write_sbml_model(rep_case_solver.model, "iML1515_generalized.xml")

def simulate_ecoli_granulate():
    model = cobra.io.read_sbml_model("iML1515_generalized.xml")
    rep_case_solver = SimpleCaseSolver(model, "BiGG")
    rep_case_solver.swap_compound("cpd11669","cpd15290")
    rep_case_solver.swap_compound("cpd11451", "cpd15500")
    cobra.io.write_sbml_model(rep_case_solver.model, "iML1515_granulated.xml")

def simulate_ecoli_granulate_boimmg():
    model = cobra.io.read_sbml_model("iML1515_generalized.xml")
    rep_case_solver = SimpleCaseSolver(model, "BiGG")
    rep_case_solver.swap_compound("cpd11669", "C_BOIMMG_749196")
    cobra.io.write_sbml_model(rep_case_solver.model, "iML1515_granulated_ubi5.xml")

def simulate_sacc_generalize():
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iMM904.xml")
    rep_case_solver = SimpleCaseSolver(model, "BiGG")
    rep_case_solver.swap_compound("cpd15290", "cpd11669")
    cobra.io.write_sbml_model(rep_case_solver.model, "iMM904_generalized.xml")

def simulate_sacc_granulate():
    model = cobra.io.read_sbml_model("iMM904_generalized.xml")
    rep_case_solver = SimpleCaseSolver(model, "BiGG")
    rep_case_solver.swap_compound("cpd11669" , "cpd15290")
    cobra.io.write_sbml_model(rep_case_solver.model, "iMM904_granulated.xml")

if __name__ == "__main__":
    # set_database_information(uri="bolt://palsson.di.uminho.pt:6094", user="neo4j", password="Tuning999")
    sacc_to_rest()
    # simulate_ecoli_generalize()
    # simulate_ecoli_granulate()
    # simulate_ecoli_granulate_boimmg()
    # simulate_sacc_generalize()
    # simulate_sacc_granulate()


from boimmgpy import definitions
from boimmgpy.id_converters.compounds_id_converter import CompoundsIDConverter
from boimmgpy.service.network_handlers.pathway_handler import printProgressBar
from boimmgpy.utilities import file_utilities
from boimmgpy.definitions import COMPOUNDS_ANNOTATION_CONFIGS_PATH


def check_if_reaction_exists(reaction_compounds, compounds_in_model_db_ids):
    """
    This method checks whether a given reaction is equal to another in the model. This is performed by removing and
    adding hydrogens to the one of the reactions.

    :param cobrapy.Reaction reaction: model reaction
    :param ModelSeedReaction reaction_container:
    :param list reaction_compounds: modelseed ids of the reactants and products
    :param list compounds_in_model_db_ids: model seed ids from the compounds in the model reaction
    :return boolean:
    """

    reaction_without_hydrogen = reaction_compounds.copy()
    reaction_with_hydrogen = reaction_compounds.copy()
    modelseed_hydrogen = "cpd00067"

    if modelseed_hydrogen in reaction_without_hydrogen:
        reaction_without_hydrogen.remove(modelseed_hydrogen)

    elif modelseed_hydrogen not in reaction_with_hydrogen:
        reaction_with_hydrogen.append(modelseed_hydrogen)

    if sorted(reaction_without_hydrogen) == sorted(compounds_in_model_db_ids) \
            or sorted(reaction_with_hydrogen) == sorted(compounds_in_model_db_ids) \
            or sorted(reaction_compounds) == sorted(compounds_in_model_db_ids):

        return True

    else:
        return False


def compare_models_reactions(model1, model2, model_database, compoundsAnnotationConfigs, compoundsIDConverter,flag=True):

    rxns1 = model1.reactions
    rxns2 = list(model2.reactions)
    reactions = rxns2.copy()
    diff_reactions_list = []
    equal_reactions_list2 = []
    equal_reactions_list1 = []
    equal_reactions = 0
    t=0
    for r1 in rxns1:
        t+=1
        printProgressBar(t,len(rxns1))
        modelseedids1 = None
        biggids1 = None
        if "seed.reaction" in r1.annotation.keys():

            modelseedids1 = r1.annotation.get("seed.reaction")
            if type(modelseedids1) == str:
                modelseedids1 = [modelseedids1]

        if "bigg.reaction" in r1.annotation.keys():
            biggids1 = r1.annotation.get("bigg.reaction")
            if type(biggids1) == str:
                biggids1 = [biggids1]

        found = False
        i = 0
        len_1 = len(r1.metabolites)

        while not found and i < len(reactions):

            r2 = reactions[i]

            len_2 = len(r2.metabolites)

            if len_1-len_2 > 1 or len_1-len_2 < -1:
                pass

            else:

                if r1.id == r2.id:
                    found = True

                if flag:
                    if not found and "seed.reaction" in r2.annotation.keys() and modelseedids1:
                        found = check_by_model_seed_reaction(r2, modelseedids1)

                    elif not found and "bigg.reaction" in r2.annotation.keys() and biggids1:
                        found = check_by_bigg_reactions(r2, biggids1)

                    if not found and r1.annotation.get("sbo") != "SBO:0000629" and r2.annotation.get("sbo") != "SBO:0000629":
                        equal = \
                            check_reaction_by_compounds(r1, r2, model_database, compoundsAnnotationConfigs,
                                                        compoundsIDConverter)

                        if equal:

                            found = True

                    if found:
                        reactions.pop(i)

            i += 1

        if found:
            equal_reactions_list2.append(r2.id)
            equal_reactions_list1.append(r1.id)
            equal_reactions += 1

        else:
            diff_reactions_list.append(r1.id)



    import matplotlib.pyplot as plt
    from matplotlib_venn import venn2

    # First way to call the 2 group Venn diagram:
    venn = venn2(subsets=(len(rxns1) - equal_reactions, len(rxns2) - equal_reactions, equal_reactions),
          set_labels=(model1.id, model2.id))

    for text in venn.set_labels:
        text.set_fontsize(14)

    for text in venn.subset_labels:
        text.set_fontsize(16)
    plt.show()

    return equal_reactions_list1,equal_reactions_list2


def check_by_model_seed_reaction(r2, modelseedids1):
    modelseedids2 = r2.annotation.get("seed.reaction")
    if type(modelseedids2) == str:
        modelseedids2 = [modelseedids2]
    j = 0
    while j < len(modelseedids1):
        l = 0
        while l < len(modelseedids2):
            if modelseedids2[l] == modelseedids1[j]:
                return True
            else:
                l += 1
        j += 1

    return False


def check_reaction_by_compounds(r1, r2, model_database, compoundsAnnotationConfigs, compoundsIDConverter):
    ids1 = get_database_ids_compounds_in_model(r1.metabolites, model_database, compoundsAnnotationConfigs,
                                               compoundsIDConverter)

    ids2 = get_database_ids_compounds_in_model(r2.metabolites, model_database, compoundsAnnotationConfigs,
                                               compoundsIDConverter)

    for id1 in ids1:
        for id2 in ids2:


            equal = check_if_reaction_exists(id2, id1)

            if equal:
                return True

    return False


def get_database_ids_compounds_in_model(compounds_in_model, model_database, compoundsAnnotationConfigs,
                                        compoundsIDConverter):
    """
    This method searches in a given database the ids of the compounds in the model

    :param dict compounds_in_model: model compounds assigned to a model reaction

    :returns list: list with database ids of the metabolites assigned to a model reaction

    """

    res = []
    for compound in compounds_in_model.keys():
        modelseed_id_annotation = compound.annotation.get("seed.compound")

        modelseed_id = []

        if modelseed_id_annotation and type(modelseed_id_annotation) == str:
            modelseed_id.append(modelseed_id_annotation)

        elif modelseed_id_annotation:
            modelseed_id.extend(modelseed_id_annotation)

        boimmg = False
        if "boimmg.compound" in compound.annotation.keys():
            boimmg_id = compound.annotation["boimmg.compound"]
            boimmg = True

        elif model_database != "ModelSEED":
            db_id = compound.annotation.get(compoundsAnnotationConfigs[model_database])
            modelseed_ids = []
            if type(db_id) == list:

                i = 0
                found = False

                while not found and i < len(db_id):
                    modelseed_ids = compoundsIDConverter.convert_dbID_into_modelSeedId(model_database,
                                                                                      db_id[i])

                    if modelseed_ids:
                        found = True
            else:
                modelseed_ids = compoundsIDConverter.convert_dbID_into_modelSeedId(model_database,
                                                                                  db_id)

            if modelseed_ids:
                if modelseed_ids[0] not in modelseed_id:
                    modelseed_id.append(modelseed_ids[0])



        if not modelseed_id and not boimmg:
            return []

        else:
            if not res:
                if boimmg:
                    res.append([boimmg_id])

                elif modelseed_id:
                    for id in modelseed_id:
                        res.append([id])

            else:

                if boimmg:
                    if modelseed_id:
                        modelseed_id.append(boimmg_id)

                    else:
                        previous = res.copy()
                        res = []
                        for reaction in previous:
                            new_reaction = reaction.copy()
                            new_reaction.append(boimmg_id)
                            res.append(new_reaction)

                if modelseed_id:
                    if len(modelseed_id) == 1:
                        for reaction in res:
                            reaction.append(modelseed_id[0])
                    else:
                        previous = res.copy()
                        res = []
                        for id in modelseed_id:
                            for reaction in previous:
                                new_reaction = reaction.copy()
                                new_reaction.append(id)
                                res.append(new_reaction)


        # elif res:
        #     for reaction in res:
        #         reaction.append(modelseed_id)
        # else:
        #     res.append([db_id])

    final_res = []
    for reaction in res:
        final_res.append(sorted(reaction))
    return final_res


def check_by_bigg_reactions(r2, biggids1):
    biggids2 = r2.annotation.get("bigg.reaction")
    if type(biggids2) == str:
        biggids2 = [biggids2]
    j = 0
    while j < len(biggids1):
        l = 0
        while l < len(biggids2):
            if biggids1[j] == biggids2[l]:
                return True
            else:
                l += 1
        j += 1

    return False


def check_by_model_seed_compounds(r2, modelseedids1):
    modelseedids2 = r2.annotation.get("seed.compound")
    if type(modelseedids2) == str:
        modelseedids2 = [modelseedids2]
    j = 0
    while j < len(modelseedids1):
        l = 0
        while l < len(modelseedids2):
            if modelseedids2[l] == modelseedids1[j]:
                return True,modelseedids2[l]
            else:
                l += 1
        j += 1

    return False,None


def check_by_bigg_compounds(r2, biggids1):
    biggids2 = r2.annotation.get("bigg.metabolite")
    if type(biggids2) == str:
        biggids2 = [biggids2]
    j = 0
    while j < len(biggids1):
        l = 0
        while l < len(biggids2):
            if biggids1[j] == biggids2[l]:
                return True,biggids1[j]
            else:
                l += 1
        j += 1

    return False,None


def compare_models_metabolites(model1, model2, met_print =False):
    mtx1 = model1.metabolites
    mtx2 = list(model2.metabolites)

    equal_reactions = 0
    model_seed_seen = []
    bigg_seen = []
    boimmg_seen = []
    ids_seen = []
    found_metabolites = []
    for r1 in mtx1:
        modelseedids1 = None
        biggids1 = None

        boimmg = False
        if "boimmg.compound" in r1.annotation.keys():
            boimmg_id = r1.annotation.get("boimmg.compound")
            boimmg = True


        if "seed.compound" in r1.annotation.keys():

            modelseedids1 = r1.annotation.get("seed.compound")
            if type(modelseedids1) == str:
                modelseedids1 = [modelseedids1]

        if "bigg.metabolite" in r1.annotation.keys():
            biggids1 = r1.annotation.get("bigg.metabolite")
            if type(biggids1) == str:
                biggids1 = [biggids1]

        found = False
        i = 0
        while not found and i < len(mtx2):
            r2 = mtx2[i]
            if r1.id == r2.id and r2.id not in ids_seen:
                ids_seen.append(r1.id)
                found = True


            # if not found and "seed.compound" in r2.annotation.keys() and modelseedids1 and r1.compartment == r2.compartment:
            #     found = check_by_model_seed_compounds(r2, modelseedids1)
            #
            # if not found and "bigg.metabolite" in r2.annotation.keys() and biggids1 and r1.compartment == r2.compartment:
            #     found = check_by_bigg_compounds(r2, biggids1)
            #
            # if not found and boimmg and "boimmg.compound" in r2.annotation.keys():
            #     boimmg_compound2 = r2.annotation.get("boimmg.compound")
            #     if boimmg_id == boimmg_compound2 and r1.compartment == r2.compartment:
            #         found = True

            if "seed.compound" in r2.annotation.keys() and modelseedids1:
                found, ms_id = check_by_model_seed_compounds(r2, modelseedids1)
                if found:
                    if ms_id not in model_seed_seen:
                        model_seed_seen.append(ms_id)
                    else:
                        found = False

            if "bigg.metabolite" in r2.annotation.keys() and biggids1:
                found,bigg_id = check_by_bigg_compounds(r2, biggids1)
                if found:
                    if bigg_id not in bigg_seen:
                        bigg_seen.append(bigg_id)
                    else:
                        found = False

            if boimmg and "boimmg.compound" in r2.annotation.keys():
                boimmg_compound2 = r2.annotation.get("boimmg.compound")
                if boimmg_id == boimmg_compound2 and boimmg_id not in boimmg_seen:
                    boimmg_seen.append(boimmg_id)
                    found = True

            i += 1

        if found:
            found_metabolites.append(r2.id)
            equal_reactions += 1

        elif met_print:

            print(r1)




    import matplotlib.pyplot as plt
    from matplotlib_venn import venn2


    # First way to call the 2 group Venn diagram:
    venn = venn2(subsets=(len(mtx1) - equal_reactions, len(mtx2) - equal_reactions, equal_reactions),
          set_labels=(model1.id, model2.id), set_colors=("b","g"),alpha=0.4)

    for text in venn.set_labels:
        text.set_fontsize(14)

    for text in venn.subset_labels:
        text.set_fontsize(16)
    plt.show()

    return found_metabolites




def compare_models(model1, model2, model3, model_database, compoundsAnnotationConfigs,
                                                 compoundsIDConverter):
    model3 = cobra.io.read_sbml_model(model3)
    model2 = cobra.io.read_sbml_model(
        model2)
    model1 = cobra.io.read_sbml_model(model1)

    found_reactions_1_1, found_reactions_1_2 = compare_models_reactions(model3, model1, model_database,
                                                                        compoundsAnnotationConfigs,
                                                                        compoundsIDConverter)
    found_1 = compare_models_metabolites(model3, model1)

    found_reactions_2_2, found_reactions_2_2 = compare_models_reactions(model2, model1, model_database,
                                                                        compoundsAnnotationConfigs,
                                                                        compoundsIDConverter)
    found_2 = compare_models_metabolites(model2, model1)

    print("--------------------- Metabolites ---------------------------")
    new_list = []
    for met2 in found_2:
        if met2 not in found_1:
            new_list.append(met2)

    for met2 in sorted(new_list):
        print(met2)

    print("--------------------- Reactions ---------------------------")
    for reaction2 in found_reactions_2_2:
        if reaction2 not in found_reactions_1_2:
            print(reaction2)

def compare_lipids_ecoli_granulated(model_database, compoundsAnnotationConfigs,
                                                 compoundsIDConverter):

    model3 = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iJR904.xml")
    model2 = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/case_studies/enhanced_model_ecoli_without_components_gap_filled.xml")
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iAF1260b_mapped.xml")




    found_reactions_1_1,found_reactions_1_2 = compare_models_reactions(model3, model, model_database, compoundsAnnotationConfigs,
                                                 compoundsIDConverter)
    found_1 = compare_models_metabolites(model3, model)

    found_reactions_2_2,found_reactions_2_2 = compare_models_reactions(model2, model, model_database, compoundsAnnotationConfigs,
                                                 compoundsIDConverter)
    found_2 = compare_models_metabolites(model2, model)

    print("--------------------- Metabolites ---------------------------")
    new_list = []
    for met2 in found_2:
        if met2 not in found_1:
            new_list.append(met2)

    for met2 in sorted(new_list):
        print(met2)

    print("--------------------- Reactions ---------------------------")
    for reaction2 in found_reactions_2_2:
        if reaction2 not in found_reactions_1_2:
            print(reaction2)



def granulated_ecoli_models(model_database, compoundsAnnotationConfigs, compoundsIDConverter):
    import cobra
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iMM904.xml")
    model2 = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/case_studies/iML1515_granulated.xml")
    model3 = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iML1515.xml")

    # compare_models_metabolites(model2, model3, met_print = True)

    # print(len(model2.reactions))
    # print(len(model3.reactions))



    found_reactions_1 = compare_models_reactions(model, model3, model_database, compoundsAnnotationConfigs, compoundsIDConverter)
    found_1 =  compare_models_metabolites(model, model3)

    found_reactions_2 = compare_models_reactions(model,model2 , model_database, compoundsAnnotationConfigs, compoundsIDConverter)
    found_2 = compare_models_metabolites(model, model2)

    print(len(found_1))
    print(len(found_2))
    for met2 in found_2:
        if met2 not in found_1:
            print(met2)

    for reaction2 in found_reactions_2:
        if reaction2 not in found_reactions_1:
            print(reaction2)



if __name__ == "__main__":
    import cobra

    compoundsIDConverter = CompoundsIDConverter()
    compoundsAnnotationConfigs = file_utilities.read_conf_file(
        COMPOUNDS_ANNOTATION_CONFIGS_PATH)
    model_database = "BiGG"

    model1 = definitions.ROOT_DIR + "/case_studies/redundant_representation_case/iML1515_mapped.xml"
    model3 = definitions.ROOT_DIR + "/case_studies/redundant_representation_case/iJR904_mapped.xml"
    model2 = definitions.ROOT_DIR + "/case_studies/redundant_representation_case/granulated_gap_filled_iJR904.xml"

    # compare_lipids_ecoli_granulated(model_database, compoundsAnnotationConfigs, compoundsIDConverter)
    compare_models(model1,model2,model3,model_database, compoundsAnnotationConfigs, compoundsIDConverter)



import itertools

from biocyc import biocyc
from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem, MolToInchiKey, MolToSmiles, rdmolops
from rdkit.Chem.rdmolfiles import MolFromSmarts, MolFromSmiles

from boimmgpy.database.databases_babel import ModelSEEDDatabases
from boimmgpy.database.accessors.compounds_database_accessor import CompoundsDBAccessor
from boimmgpy.service.network_handlers.pathway_handler import PathwayHandler
from boimmgpy.id_converters.compounds_id_converter import CompoundsIDConverter
from boimmgpy.id_converters.reactions_id_converter import ReactionsIDConverter
from boimmgpy.model_seed.model_seed_compounds_database import ModelSeedCompoundsDB
from boimmgpy.model_seed.model_seed_reactions_database import ModelSeedReactionsDB
from boimmgpy.utilities import chemo_utilities, file_utilities
from boimmgpy.definitions import EXCEPTIONS


def printProgressBar(iteration, total, prefix='', suffix='', decimals=1, length=100, fill='|', printEnd="\r"):
    """
  Call in a loop to create terminal progress bar
  @params:
      iteration   - Required  : current iteration (Int)
      total       - Required  : total iterations (Int)
      prefix      - Optional  : prefix string (Str)
      suffix      - Optional  : suffix string (Str)
      decimals    - Optional  : positive number of decimals in percent complete (Int)
      length      - Optional  : character length of bar (Int)
      fill        - Optional  : bar fill character (Str)
      printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
  """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print('\r%s |%s| %s%% %s' % (prefix, bar, percent, suffix), end="",flush=True)
    # Print New Line on Complete
    if iteration == total:
        print()


class TransformationsHandler:

    def __init__(self):
        biocyc.set_organism('meta')
        self.exceptions_conf = file_utilities.read_conf_file(EXCEPTIONS)
        self.reactions_transformations = {}

        self.compounds_simplifier = {"cpd24471": "cpd12300",
                                     "cpd27426": "cpd27914",
                                     "cpd11611": "cpd29330",
                                     "cpd00487": "cpd19000"}

    @property
    def reactions_transformations(self):
        return self.__reactions_transformations

    @reactions_transformations.setter
    def reactions_transformations(self, reactions_transformations):
        self.__reactions_transformations = reactions_transformations

    @property
    def modelSeedCompoundsDB(self):
        return self.__modelSeedCompDB

    @modelSeedCompoundsDB.setter
    def modelSeedCompoundsDB(self, db):
        self.__modelSeedCompDB = db

    @property
    def compoundsIDConverter(self):
        return self.__compoundsIDConverter

    @compoundsIDConverter.setter
    def compoundsIDConverter(self, compoundsIDConverter):
        self.__compoundsIDConverter = compoundsIDConverter

    def extract_metacyc_pathways(self, parent, generic=False):

        biocyc.set_organism('meta')
        ptw_class = biocyc.get(parent)
        l = [ptw_class]
        res = []
        while len(l) > 0:
            try:
                node = l.pop(0)

                if not node.subclasses and not node.instances:
                    res.append(node)
                    # print(node.name)

                following = node.subclasses
                following.extend(node.instances)
                for elem in following:
                    if elem not in res and elem not in l:
                        l.insert(0, elem)

            except:
                pass

        i = 0
        for instance in res:
            try:
                # printProgressBar(i, len(res))
                self.__pathway_handler.add_metacyc_pathway(instance.id, generic)
                i += 1

            except:
                pass

    def __simplify_compounds(self, modelseed_compounds):
        res = []
        for compound in modelseed_compounds:
            if compound in self.compounds_simplifier.keys():
                res.append(self.compounds_simplifier.get(compound))
            else:
                res.append(compound)

        return res

    def choose_transformations_from_pathway(self,pathway_parent,compounds_to_avoid = [], accept_cycles = False):

        self.modelSeedCompoundsDB = ModelSeedCompoundsDB()
        self.modelSeedReactionsDB = ModelSeedReactionsDB()
        self.reactionsIDConverter = ReactionsIDConverter()

        self.__pathway_handler = PathwayHandler(self.modelSeedCompoundsDB)

        self.extract_metacyc_pathways(pathway_parent, True)
        target_pathways = self.__pathway_handler.get_target_pathways()


        for pathway in target_pathways:

            paths_within_pathway = self.__pathway_handler.get_all_paths_using_pathway_id(pathway,accept_cycles)
            for path in paths_within_pathway:

                reverse_pathway = path[::-1]

                for reaction in reverse_pathway:

                    print(reaction)

                    if reaction not in self.reactions_transformations.keys():

                        reactions = self.reactionsIDConverter.convert_db_id_to_model_seed_by_db_id(reaction)

                        found=False
                        for react in reactions:
                            temp_reaction = self.modelSeedReactionsDB.getReaction(react)
                            compounds = temp_reaction.getCompounds()
                            for compound in compounds:
                                temp_compound = self.modelSeedCompoundsDB.get_compound_by_id(compound)
                                if "R" in temp_compound.getFormula() or "*" in temp_compound.getSmiles():
                                    model_seed_reaction = temp_reaction
                                    found = True
                                    break
                            if found:
                                break

                        # model_seed_reaction = self.modelSeedReactionsDB.getReaction(reaction_id[0])
                        if found:
                            reactants = model_seed_reaction.getReactants()
                            if "cpd26685" in reactants:
                                reactants.remove("cpd26685")
                                reactants.append("cpd00167")

                            if "cpd28189" in reactants:
                                reactants.remove("cpd28189")
                                reactants.append("cpd00431")

                            products = model_seed_reaction.getProducts()
                            if "cpd26685" in products:
                                products.remove("cpd26685")
                                products.append("cpd00167")

                            if "cpd28189" in products:
                                products.remove("cpd28189")
                                products.append("cpd00431")

                            reactants = self.__simplify_compounds(reactants)
                            products = self.__simplify_compounds(products)

                            model_seed_reactants = []

                            for reac in reactants:
                                model_seed_reactant = self.modelSeedCompoundsDB.get_compound_by_id(reac)

                                if "R" in model_seed_reactant.getFormula() or "*" in model_seed_reactant.getSmiles():

                                    if (model_seed_reactant.getSmiles() or
                                        model_seed_reactant.getDbId() in self.exceptions_conf.keys())  and \
                                            model_seed_reactant.getDbId() not in compounds_to_avoid:

                                        model_seed_reactants.append(model_seed_reactant)

                            model_seed_products = []
                            for prod in products:

                                model_seed_product = self.modelSeedCompoundsDB.get_compound_by_id(prod)

                                if "R" in model_seed_product.getFormula() or "*" in model_seed_product.getSmiles():

                                    if (model_seed_product.getSmiles() or
                                        model_seed_product.getDbId() in self.exceptions_conf.keys())  and \
                                        model_seed_product.getDbId() not in compounds_to_avoid:

                                        model_seed_products.append(model_seed_product)



                        # products_model_seed = self.__convert_metacyc_compound_into_model_seed(products)
                        # reactants_model_seed = self.__convert_metacyc_compound_into_model_seed(reactants)

                            if model_seed_products:

                                transformation, _ = self.__convert_reaction_into_transformation(model_seed_reactants,
                                                                                            model_seed_products)

                                self.reactions_transformations[reaction] = transformation


    def save_transformations(self,filename):

        with open(filename,"w") as file:

            for reaction in self.reactions_transformations.keys():
                file.write(reaction+"!!!"+self.reactions_transformations[reaction]+"\n")

    def load_transformations(self,filepath):

        with open(filepath,"r") as file:

            lines = file.readlines()

            for line in lines:
                if line:
                    line_lst = line.split("!!!")
                    reaction = line_lst[0]
                    transformation = line_lst[1].strip()

                    self.reactions_transformations[reaction] = transformation





    def __convert_metacyc_compound_into_model_seed(self, metacyc_compounds):
        res = []
        for metacyc_compound in metacyc_compounds:
            modelseedids = self.compoundsIDConverter.convert_dbID_into_modelSeedId("MetaCyc", metacyc_compound.id)
            if modelseedids:
                model_seed_compound = self.__check_if_generic(modelseedids)
                if model_seed_compound:
                    res.append(model_seed_compound)

        return res


    def __neutralize_molecules(self, molecules):
        molecules_smiles = []

        for molecule in molecules:
            model_seed_id = molecule.getDbId()
            if model_seed_id in self.exceptions_conf.keys():
                new_smiles = self.exceptions_conf.get(model_seed_id)
            else:
                new_smiles, neutralised = chemo_utilities.NeutraliseCharges(molecule.getSmiles())

            molecules_smiles.append(new_smiles)

        return molecules_smiles


    def __convert_reaction_into_transformation(self, reactants, products):
        """
        This method aims at transforming a MetaCyc reaction into a SMIRK transformation

        :param reactants: list of modelseed generic reactants
        :param products: list of modelseed generic products
        :return: a string representing the transformation
        """



        reactants_smiles = self.__neutralize_molecules(reactants)
        products_smiles = self.__neutralize_molecules(products)

        react = ".".join(reactants_smiles)
        prod = ".".join(products_smiles)

        r_groups_reactants = react.count("*")
        r_groups_products = prod.count("*")

        n_primary_precursors = 0
        while r_groups_reactants<r_groups_products:
            reactants_smiles.append(reactants_smiles[0]) #martelado

            react = ".".join(reactants_smiles)
            r_groups_reactants = react.count("*")
            n_primary_precursors += 1

        reactant_smiles_combinations = chemo_utilities.get_combinations_of_dummy_atoms(react)
        product_smiles = chemo_utilities.convert_model_seed_dummy_atoms(prod)

        transformations_combinations = []
        for reactant_smiles in reactant_smiles_combinations:
            reaction_smart = product_smiles + ">>" + reactant_smiles
            transformations_combinations.append(reaction_smart)

        user_interaction_string = ""
        i = 1
        dict_for_user_to_choose = {}

        if len(transformations_combinations) > 1:

            for transformation in transformations_combinations:
                user_interaction_string += str(i) + ": "
                user_interaction_string += transformation + "\n"
                dict_for_user_to_choose[i] = transformation
                i += 1

            transformation_i = input(user_interaction_string)
            chosen_transformation = dict_for_user_to_choose[int(transformation_i)]

        else:
            chosen_transformation = transformations_combinations[0]


        return chosen_transformation,n_primary_precursors

    def __check_if_generic(self, modelseedids):
        for modelseedid in modelseedids:

            compound = self.modelSeedCompoundsDB.get_compound_by_id(modelseedid)
            formula = compound.getFormula()
            if "R" in formula:
                return compound
        return None


class OntologyGeneratorDA:

    def __init__(self):

        self.cores = None
        self.modelSeedCompoundsDB = ModelSeedCompoundsDB()
        self.modelSeedReactionsDB = ModelSeedReactionsDB()
        biocyc.set_organism("meta")
        # self.__modelSeedCompoundsDatabase = ModelSeedCompoundsDB()
        self.accessor = CompoundsDBAccessor()

        self.compoundsIDConverter = CompoundsIDConverter()
        self.reactionsIDConverter = ReactionsIDConverter()

        self.compounds_simplifier = {"cpd24471":"cpd12300",
                                     "cpd27426":"cpd27914",
                                     "cpd11611":"cpd29330",
                                     "cpd00487":"cpd19000"}


        ############################## exceptionally for biosynthetic pathways in the sphingolipids pathway ##################


        self.compounds_simplifier["cpd26685"] = "cpd00167"
        self.compounds_simplifier["cpd28189"] = "cpd00431"

        ######################################## ###############################################################################


        self.exceptions_conf = file_utilities.read_conf_file(EXCEPTIONS)

    @property
    def modelSeedCompoundsDB(self):
        return self.__modelSeedCompDB

    @modelSeedCompoundsDB.setter
    def modelSeedCompoundsDB(self,db):
        self.__modelSeedCompDB = db

    @property
    def modelSeedReactionsDB(self):
        return self.__modelSeedReactionsDB

    @modelSeedReactionsDB.setter
    def modelSeedReactionsDB(self, db):
        self.__modelSeedReactionsDB = db

    @property
    def transformationsHandler(self):
        return self.__transformationsHandler

    @transformationsHandler.setter
    def transformationsHandler(self, transformationsHandler):
        self.__transformationsHandler = transformationsHandler

    @property
    def cores(self):
        return self.__cores

    @cores.setter
    def cores(self, core):
        self.__cores = core

    @property
    def compoundsIDConverter(self):
        return self.__compoundsIDConverter

    @compoundsIDConverter.setter
    def compoundsIDConverter(self,compoundsIDConverter):
        self.__compoundsIDConverter = compoundsIDConverter

    @property
    def reactionsIDConverter(self):
        return self.__reactionsIDConverter

    @reactionsIDConverter.setter
    def reactionsIDConverter(self, reactionsIDConverter):
        self.__reactionsIDConverter = reactionsIDConverter

    @property
    def accessor(self):
        return self.__accessor

    @accessor.setter
    def accessor(self,accessor):
        self.__accessor = accessor

    @property
    def compounds_simplifier(self):
        return self.__compounds_simplifier

    @compounds_simplifier.setter
    def compounds_simplifier(self,value):
        self.__compounds_simplifier = value

    def get_combinations_with_the_same_components(self,combinations):
        res = []
        i=0

        comp_and_components = {}
        print("Getting combinations... ")
        for combination in combinations:
            i+=1

            printProgressBar(i,len(combinations))

            if combination[0] not in comp_and_components:
                temp_components = self.accessor.get_predecessors_by_ont_id_rel_type(combination[0], "component_of")
                comp_and_components[combination[0]] = temp_components

            else:
                temp_components = comp_and_components[combination[0]]

            found = False
            for compound in combination[1:]:
                if compound not in comp_and_components.keys():
                    components = self.accessor.get_predecessors_by_ont_id_rel_type(compound,"component_of")
                    comp_and_components[compound] = components

                else:
                    components = comp_and_components[compound]

                if components:
                    component = components[0]
                    if component in temp_components:
                        found = True
                    else:
                        found = False
                        break
            if found:
                res.append(combination)
        return res


    def create_new_compounds_using_transformation(self,transformation,reactants, products,reaction_id="",
                                                  source="",add_rel= False,same_components = False,
                                                  reverse=True):
        children = []
        for reactants_ont_id in reactants:

            if same_components:
                reactant_children = self.accessor.get_predecessors_with_same_component(reactants_ont_id,"is_a")

            else:
                reactant_children = self.accessor.get_predecessors_by_ont_id_rel_type(reactants_ont_id,"is_a")

            children.append(reactant_children.copy())

        combinations = list(itertools.product(*children))
        if same_components and len(children) > 1:
            combinations = self.get_combinations_with_the_same_components(combinations)


        i = 0
        for combination in combinations:
            printProgressBar(i + 1, len(combinations), prefix='Progress:', suffix='Complete')
            i += 1

            new_molecules_list = self.generate_new_molecule_with_transformation(list(combination),
                                                                                transformation)

            for new_molecule in new_molecules_list:
                self.__add_molecule_and_relationships(products,
                                                      new_molecule,
                                                      reaction_id,
                                                      source = source,
                                                      add_biosynthetic_rel = add_rel,reverse= reverse)


    def handle_manual_reaction(self,reaction_transformation,reactants,products,reaction_id,source):

        model_seed_products = []
        for product in products:
            model_seed_compound = self.modelSeedCompoundsDB.get_compound_by_id(product)
            model_seed_products.append(model_seed_compound)

        # products_model_seed = self.__convert_metacyc_compounds_into_model_seed(products)

        if model_seed_products:

            products_ont_ids = self.__get_generic_compounds_from_ontology(model_seed_products)

            children = []
            for products_ont_id in products_ont_ids:
                product_children = self.accessor.get_predecessors_by_ont_id_rel_type(products_ont_id,"is_a")
                children.append(product_children)

            combinations = list(itertools.product(*children))

            i = 0
            for combination in combinations:

                printProgressBar(i + 1, len(children), prefix='Progress:', suffix='Complete')
                i += 1

                reactants = [self.modelSeedCompoundsDB.get_compound_by_id(reac) for reac in reactants]
                products = [self.modelSeedCompoundsDB.get_compound_by_id(prod) for prod in products]

                new_reaction_products = self.__check_if_product_is_compatible_with_ont_container(products,
                                                                                                 combination)

                reactants_ont_ids = self.__get_generic_compounds_from_ontology(reactants)

                for reactants_ont_id in reactants_ont_ids:
                    for products_ont_id in products_ont_ids:
                        self.accessor.add_biosynthetic_relationship(reactants_ont_id, products_ont_id,
                                                                        reaction=reaction_id)


                if reactants_ont_ids and new_reaction_products:

                    new_molecules_list = self.generate_new_molecule_with_transformation(new_reaction_products,
                                                                                        reaction_transformation)


                    for new_molecules in new_molecules_list:

                        if new_molecules:

                            self.__add_molecule_and_relationships(reactants_ont_ids,
                                                                  new_molecules,
                                                                  reaction_id,
                                                                  source)

    def handle_reaction(self,reaction, pathway_id, transformations_file):
        self.transformationsHandler = TransformationsHandler()
        self.transformationsHandler.load_transformations(transformations_file)

        modelseedid = self.reactionsIDConverter.convert_db_id_to_model_seed_by_db_id(reaction)
        model_seed_reaction = self.modelSeedReactionsDB.getReaction(modelseedid[0])

        # products = metacyc_reaction.compounds_right
        products = model_seed_reaction.getProducts()
        model_seed_products = []

        for product in products:
            model_seed_compound = self.modelSeedCompoundsDB.get_compound_by_id(product)
            model_seed_products.append(model_seed_compound)

        # products_model_seed = self.__convert_metacyc_compounds_into_model_seed(products)

        if model_seed_products:

            products_ont_ids = self.__get_generic_compounds_from_ontology(model_seed_products)

            children = []
            for products_ont_id in products_ont_ids:

                product_children = self.accessor.get_predecessors_by_ont_id_rel_type(products_ont_id,"is_a")

                children.append(product_children)

            combinations = list(itertools.product(*children))

            i = 0
            for combination in combinations:

                printProgressBar(i + 1, len(children), prefix='Progress:', suffix='Complete')
                i += 1

                reaction_id = self.reactionsIDConverter.convert_db_id_to_model_seed_by_db_id(reaction)
                model_seed_reaction = self.modelSeedReactionsDB.getReaction(reaction_id[0])
                model_seed_reactants = model_seed_reaction.getReactants()
                model_seed_products = model_seed_reaction.getProducts()

                model_seed_reactants = self.__simplify_compounds(model_seed_reactants)
                model_seed_products = self.__simplify_compounds(model_seed_products)

                reactants = [self.modelSeedCompoundsDB.get_compound_by_id(reac) for reac in model_seed_reactants]
                products = [self.modelSeedCompoundsDB.get_compound_by_id(prod) for prod in model_seed_products]

                new_reaction_products = self.__check_if_product_is_compatible_with_ont_container(products,
                                                                                                 combination)

                reactants_ont_ids = self.__get_generic_compounds_from_ontology(reactants)

                for reactants_ont_id in reactants_ont_ids:
                    for products_ont_id in products_ont_ids:
                        self.accessor.add_biosynthetic_relationship(reactants_ont_id, products_ont_id,
                                                                        reaction=reaction,
                                                                        pathway=pathway_id)


                if reactants_ont_ids and new_reaction_products:
                    transformation = self.transformationsHandler.reactions_transformations[reaction]

                    new_molecules_list = self.generate_new_molecule_with_transformation(new_reaction_products,
                                                                                        transformation)


                    for new_molecules in new_molecules_list:

                        if new_molecules:

                            self.__add_molecule_and_relationships(reactants_ont_ids,
                                                                  new_molecules,
                                                                  reaction,
                                                                  pathway_id)




    def __call__(self,pathway_parent,cores,transformations_file,pathway_to_avoid = [], reactions_to_avoid = [],
                 accept_cycles = False,list_of_targets = [],all=True):

        self.transformationsHandler = TransformationsHandler()
        self.transformationsHandler.load_transformations(transformations_file)
        self.seen = pathway_to_avoid

        self.reactions_to_avoid = reactions_to_avoid
        self.cores = cores
        self.__pathway_handler = PathwayHandler(self.modelSeedCompoundsDB)
        self.extract_metacyc_pathways(pathway_parent,True)
        target_pathways = self.__pathway_handler.get_target_pathways()


        for pathway in target_pathways:
            print("Trying to handle this pathway: %s"%pathway)
            if pathway not in self.seen:
                paths_within_pathways = self.__pathway_handler.get_all_paths_using_pathway_id(pathway,accept_cycles)

                for path in paths_within_pathways:
                    to_continue = True
                    for reaction in self.reactions_to_avoid:
                        if reaction in path:
                            to_continue=False
                            break

                    if to_continue:
                        reverse_path = path[::-1]
                        print("Target: %s"%reverse_path[0])
                        self.__handle_pathway(reverse_path,pathway,list_of_targets,all)

                # self.seen.extend(reverse_pathway)



    def __handle_pathway(self,reverse_pathway,pathway_id,list_of_targets=[],all=True):


        print("Handling the following pathway: %s"%pathway_id)
        first_reaction = reverse_pathway[0]

        # metacyc_reaction = biocyc.get(first_reaction)
        modelseedid = self.reactionsIDConverter.convert_db_id_to_model_seed_by_db_id(first_reaction)
        model_seed_reaction = self.modelSeedReactionsDB.getReaction(modelseedid[0])

        # products = metacyc_reaction.compounds_right
        products = model_seed_reaction.getProducts()
        model_seed_products = []

        for product in products:
            model_seed_compound = self.modelSeedCompoundsDB.get_compound_by_id(product)
            model_seed_products.append(model_seed_compound)


        # products_model_seed = self.__convert_metacyc_compounds_into_model_seed(products)

        if model_seed_products:

            products_ont_ids = self.__get_generic_compounds_from_ontology(model_seed_products)

            i = 0

            if not list_of_targets:
                children = []

                for products_ont_id in products_ont_ids:

                    if not all:
                        product_children = self.accessor.get_predecessors_only_from_lipid_maps_and_ms(products_ont_id)

                    else:
                        product_children = self.accessor.get_predecessors_by_ont_id(products_ont_id)

                    children.append(product_children)

            else:
                children = [list_of_targets]

            combinations = list(itertools.product(*children))

            for combination in combinations:

                printProgressBar(i + 1, len(combinations), prefix='Progress:', suffix='Complete')
                i += 1

                children_products_ont_ids = list(combination)
                for reaction in reverse_pathway:

                    # metacyc_reaction = biocyc.get(reaction)
                    # print(metacyc_reaction.id)
                    # reactants = metacyc_reaction.compounds_left
                    # products = metacyc_reaction.compounds_right


                    reaction_id = self.reactionsIDConverter.convert_db_id_to_model_seed_by_db_id(reaction)
                    model_seed_reaction = self.modelSeedReactionsDB.getReaction(reaction_id[0])
                    model_seed_reactants = model_seed_reaction.getReactants()
                    model_seed_products = model_seed_reaction.getProducts()

                    model_seed_reactants = self.__simplify_compounds(model_seed_reactants)
                    model_seed_products = self.__simplify_compounds(model_seed_products)

                    reactants = [self.modelSeedCompoundsDB.get_compound_by_id(reac) for reac in model_seed_reactants]
                    products = [self.modelSeedCompoundsDB.get_compound_by_id(prod) for prod in model_seed_products]

                    if children_products_ont_ids:

                        for child in children_products_ont_ids:
                            new_reaction_products = self.__check_if_product_is_compatible_with_ont_container(products,[child])

                            # reactants_model_seed = self.__convert_metacyc_compounds_into_model_seed(reactants)
                            reactants_ont_ids = self.__get_generic_compounds_from_ontology(reactants)


                            if reaction not in self.seen:
                                for reactants_ont_id in reactants_ont_ids:
                                    for products_ont_id in products_ont_ids:

                                        self.accessor.add_biosynthetic_relationship(reactants_ont_id,products_ont_id,
                                                                                    reaction = reaction,
                                                                                    pathway=pathway_id)

                                self.seen.append(reaction)

                            if reactants_ont_ids and new_reaction_products:

                                if reaction in self.transformationsHandler.reactions_transformations:
                                    transformation = self.transformationsHandler.reactions_transformations[reaction]

                            # the following describe the reaction in a reverse way,
                            # the products of the actual reaction will be the reactants in the transformation

                                    new_molecules_list = self.generate_new_molecule_with_transformation(new_reaction_products,transformation)

                                    new_products = []

                                    for new_molecules in new_molecules_list:

                                        if new_molecules:
                                            new_molecules_ont_ids = self.__add_molecule_and_relationships(reactants_ont_ids,
                                                                                                          new_molecules,reaction,pathway_id)
                                            new_products.extend(new_molecules_ont_ids)

                                    children_products_ont_ids = new_products.copy()





    def generate_new_molecule_with_transformation(self, reactants, transformation):
        """

        :param reactants: ontology ids of the different possible reactants
        :param transformation: SMART transformation
        :return list: list of tuples (ontology_id of downstream compound, new molecules smiles)
        """
        res = []
        reactants_mol =[]
        for reactant in reactants:

            reactant = self.accessor.get_node_by_ont_id(reactant)

            if not reactant.smiles:
                return []

            neutralized_smiles, _ = chemo_utilities.NeutraliseCharges(reactant.smiles)
            reactant_mol = MolFromSmiles(neutralized_smiles)

            reactants_mol.append(reactant_mol)

        reactants_tuple = tuple(reactants_mol)


        transformation_smarts = AllChem.ReactionFromSmarts(transformation)

        try:
            new_molecules = [Chem.MolToSmiles(x, 1) for x in transformation_smarts.RunReactants(
                reactants_tuple)[0]]

            res.append((reactants,new_molecules))
        except:
            res.append(None)

        return res



    def extract_metacyc_pathways(self,parent,generic = False):

        biocyc.set_organism('meta')
        ptw_class = biocyc.get(parent)
        l = [ptw_class]
        res = []
        while len(l) > 0:
            try:
                node = l.pop(0)

                if not node.subclasses and not node.instances:
                    res.append(node)
                    # print(node.name)

                following = node.subclasses
                following.extend(node.instances)
                for elem in following:
                    if elem not in res and elem not in l:
                        l.insert(0, elem)

            except:
                pass

        i = 0
        for instance in res:
            try:
                # printProgressBar(i, len(res))
                self.__pathway_handler.add_metacyc_pathway(instance.id,generic)
                i += 1

            except:
                pass


    def get_similarity(self,molecules, reactant):
        similarities = []
        for molecule in molecules:
            fgp1 = AllChem.GetMorganFingerprint(MolFromSmiles(molecule), 3, useFeatures=True,useChirality=True)
            fgp_reactant = AllChem.GetMorganFingerprint(reactant, 1, useFeatures=True,useChirality=True)
            similarity = DataStructs.TanimotoSimilarity(fgp_reactant, fgp1,useFeatures=True,useChirality=True)


            similarities.append(similarity)


        average = sum(similarities) / len(similarities)
        return average

    def __get_filtered_compounds(self,smiles_keys,smarts,smiles_database):
        new_smiles_list = smiles_keys.copy()
        filtered_compounds = []
        for smile in smiles_keys:
            molecule = MolFromSmiles(smile)
            if molecule:
                match = molecule.GetSubstructMatch(smarts)

                if match:
                    filtered_compounds.append(smiles_database[smile])
                    new_smiles_list.remove(smile)

        smiles_keys = new_smiles_list.copy()

        return (filtered_compounds,smiles_keys)


    def __convert_metacyc_compounds_into_model_seed(self, metacyc_compounds):
        res = []
        for metacyc_compound in metacyc_compounds:
            modelseedids = self.__compoundsIDConverter.convert_dbID_into_modelSeedId("MetaCyc", metacyc_compound.id)
            if modelseedids:
                model_seed_compound = self.__check_if_generic(modelseedids)
                if model_seed_compound:
                    res.append(model_seed_compound)

        return res

    def __check_if_generic(self, modelseedids):
        for modelseedid in modelseedids:

            compound = self.modelSeedCompoundsDB.get_compound_by_id(modelseedid)
            formula = compound.getFormula()
            if "R" in formula:
                return compound
        return None

    def __neutralize_molecules(self, molecules):
        molecules_smiles = []
        for molecule in molecules:
            #mol = MolFromSmiles(molecule.getSmiles())
            new_smiles, neutralised = chemo_utilities.NeutraliseCharges(molecule.getSmiles())
            molecules_smiles.append(new_smiles)

        return molecules_smiles

    def __get_generic_compounds_from_ontology(self, model_seed_molecules):

        res = []

        for compound in model_seed_molecules:
            db_id = compound.getDbId()
            smiles = compound.getSmiles()
            smiles,_ = chemo_utilities.NeutraliseCharges(smiles)
            if "*" in smiles or "R" in compound.getFormula():
                ont_id = self.accessor.get_node_from_model_seed_id(db_id).id

                res.append(ont_id)

        return res

    def __add_molecule_and_relationships(self, possible_structural_parents, precursor_successor_and_smiles,reaction_id,
                                         source,pathway_id=None,add_biosynthetic_rel = True, reverse= True):

        new_compounds_ont_ids = []
        precursor_successors = precursor_successor_and_smiles[0]
        smiles_lst = precursor_successor_and_smiles[1]

        for smiles in smiles_lst:
            mol= MolFromSmiles(smiles)

            found = False
            i=0
            while not found and i < len(possible_structural_parents):
                structuralParentOntId = possible_structural_parents[i]
                structuralParentContainer = self.accessor.get_node_by_ont_id(structuralParentOntId)
                generic_smiles = structuralParentContainer.smiles
                isStructuralParent = \
                    chemo_utilities.check_similarity_between_generic_and_complete_representation(generic_smiles, smiles)

                if isStructuralParent:
                    found = True

                i+=1

            if found:


                newCompound_ont_id = self.__generate_complete_compound_and_establish_relationships(mol,structuralParentContainer,
                                                                          precursor_successors,
                                                                        reaction_id,pathway_id,source,
                                                                        add_biosynthetic_rel,reverse)

                if newCompound_ont_id not in new_compounds_ont_ids:
                    new_compounds_ont_ids.append(newCompound_ont_id)

        return new_compounds_ont_ids




    def __generate_complete_compound_and_establish_relationships(self,
                                                                 mol,
                                                                 structuralParentContainer,
                                                                 precursor_sucessors,reaction_id,
                                                                 pathway_id,source,add_biosynthetic_rel=True,
                                                                 reverse = True):


        inchikey = MolToInchiKey(mol)
        smiles = MolToSmiles(mol)
        if inchikey:
            node = self.accessor.get_compound_by_inchikey(inchikey)
        else:
            node = self.accessor.get_compound_by_smiles(smiles)

        if node and precursor_sucessors:
            new_molecule_ont_id = node.id

            if add_biosynthetic_rel:
                for successor in precursor_sucessors:
                    if pathway_id:
                        if reverse:
                            self.accessor.add_biosynthetic_relationship(new_molecule_ont_id, successor,
                                                            reaction=reaction_id,pathway=pathway_id,source=source)
                        else:
                            self.accessor.add_biosynthetic_relationship(successor,new_molecule_ont_id,
                                                                        reaction=reaction_id, pathway=pathway_id,
                                                                        source=source)
                    elif reverse:
                        self.accessor.add_biosynthetic_relationship(new_molecule_ont_id, successor,
                                                                reaction=reaction_id, source=source)
                    else:
                        self.accessor.add_biosynthetic_relationship(successor, new_molecule_ont_id,
                                                                    reaction=reaction_id, source=source)

            self.accessor.establish_structural_relationship(new_molecule_ont_id,structuralParentContainer.id)



        else:
            formula = Chem.rdMolDescriptors.CalcMolFormula(mol)

            if "*" in formula:
                number_r_groups = formula.count("*")
                formula = formula.replace("*","") + "R" + str(number_r_groups)

            inchikey = MolToInchiKey(mol)
            smiles = MolToSmiles(mol)

            side_chain_providers = self.__find_side_chain_providers(mol, structuralParentContainer,precursor_sucessors)

            name = self.__generate_new_molecule_name(structuralParentContainer, side_chain_providers)

            charge = rdmolops.GetFormalCharge(mol)

            new_molecule_ont_container = self.accessor.create_compound(name=name,formula=formula,
                                                                       inchikey=inchikey,
                                                                       smiles=smiles,annotated =False,charge = charge)
            new_molecule_ont_id = new_molecule_ont_container.id

            self.accessor.establish_structural_relationship(new_molecule_ont_container.id,structuralParentContainer.id)

            for side_chain_provider in side_chain_providers:
                self.accessor.add_relationship(side_chain_provider.id,new_molecule_ont_id,"component_of")

            if add_biosynthetic_rel:
                for successor in precursor_sucessors:
                    if pathway_id:
                        if reverse:
                            self.accessor.add_biosynthetic_relationship(new_molecule_ont_id, successor,
                                                                        reaction=reaction_id, pathway=pathway_id,
                                                                        source=source)
                        else:
                            self.accessor.add_biosynthetic_relationship(successor, new_molecule_ont_id,
                                                                        reaction=reaction_id, pathway=pathway_id,
                                                                        source=source)
                    elif reverse:
                        self.accessor.add_biosynthetic_relationship(new_molecule_ont_id, successor,
                                                                    reaction=reaction_id, source=source)
                    else:
                        self.accessor.add_biosynthetic_relationship(successor, new_molecule_ont_id,
                                                                    reaction=reaction_id, source=source)

            self.accessor.establish_structural_relationship(new_molecule_ont_id,structuralParentContainer.id)

        return new_molecule_ont_id

    def __generate_new_molecule_name(self,product_container,combination_list):

        names = []

        for container in combination_list:
            name = container.name
            if name not in names:
                names.append(name)

        new_name = product_container.name
        for unique_name in names:
            new_name += " (" + unique_name + ") "

        return new_name



    def __generate_complete_compound(self,lipid_compound,
                                     model_seed_compound,
                                     mol,
                                     structuralParentContainer=None,
                                     side_chain_providers=None):

        if lipid_compound:
            aliases = lipid_compound.getAliases()
            formula = lipid_compound.getFormula()
            inchikey = lipid_compound.getInchiKey()
            smiles = lipid_compound.getSmiles()
            name = lipid_compound.getName()

        if lipid_compound and model_seed_compound:
            id = model_seed_compound.getDbId()
            model_seed_aliases = self.__compoundsIDConverter.get_all_aliases_by_modelSeedID(id)
            aliases.update(model_seed_aliases)

        elif lipid_compound and not model_seed_compound:
            id = lipid_compound.getDbId()


        elif model_seed_compound:
            id = model_seed_compound.getDbId()
            model_seed_aliases = self.__compoundsIDConverter.get_all_aliases_by_modelSeedID(id)
            aliases = model_seed_aliases
            formula = model_seed_compound.getFormula()
            inchikey = model_seed_compound.getInchikey()
            smiles = model_seed_compound.getSmiles()
            name = model_seed_compound.getName()

        # elif structuralParentContainer and side_chain_providers:
        else:
            id = ""
            formula = Chem.rdMolDescriptors.CalcMolFormula(mol)
            inchikey = MolToInchiKey(mol)
            smiles = MolToSmiles(mol)
            aliases = {}
            name = self.__generate_new_molecule_name(structuralParentContainer, side_chain_providers)

        return id,name,formula,inchikey,smiles,aliases


    def __find_side_chain_providers(self,mol,structuralParentContainer,precursor_sucessors):

        r_groups_n1 = structuralParentContainer.smiles.count("*")
        r_groups_n2 = 0


        side_chains = []
        for precursor in precursor_sucessors:
            side_chains_lst = \
                self.accessor.get_predecessors_by_ont_id_rel_type(precursor,"component_of")


            # try:
            precursor_parent = self.accessor.get_successors_by_ont_id_rel_type(precursor,"is_a")
            node_container = self.accessor.get_node_by_ont_id(precursor_parent[0])
            r_groups_n2 += node_container.smiles.count("*")
            # except:
            #     pass

            side_chains.extend(side_chains_lst)

        model_seed_ids = structuralParentContainer.aliases[ModelSEEDDatabases.MODEL_SEED.value]

        if r_groups_n1 == r_groups_n2 and all(ms_id not in self.exceptions_conf for ms_id in model_seed_ids):
            res = []

            for side_chain in side_chains:
                container = self.accessor.get_node_by_ont_id(side_chain)
                res.append(container)

            return res


        parent_neutralized_smiles, _ = chemo_utilities.NeutraliseCharges(structuralParentContainer.smiles)
        backup_parent = parent_neutralized_smiles
        parent_neutralized_smiles = parent_neutralized_smiles.replace("(*)", "").replace("*", "")

        parent_smarts = MolFromSmarts(parent_neutralized_smiles)
        if not parent_smarts:
            parent_smarts = MolFromSmarts(backup_parent)


        sidechain_compound = AllChem.DeleteSubstructs(mol, parent_smarts)
        sidechain2, sidechains_smiles_list2 = chemo_utilities.retrieve_fragments(sidechain_compound)
        components = []

        for sidechain in side_chains:
            container = self.accessor.get_node_by_ont_id(sidechain)

            neutralized_smiles,_ = chemo_utilities.NeutraliseCharges(container.smiles)

            for core in self.cores:
                sidechain1 = AllChem.DeleteSubstructs(MolFromSmiles(neutralized_smiles),MolFromSmiles(core))
                sidechain1, sidechains_smiles_list1 = chemo_utilities.retrieve_fragments(sidechain1)

                for side_chain_smiles in sidechains_smiles_list2:
                    if sidechains_smiles_list1:
                        equal_side_chains = \
                            chemo_utilities.check_if_molecules_are_equal(MolFromSmiles(side_chain_smiles), MolFromSmiles(sidechains_smiles_list1[0]))

                        if equal_side_chains:
                            if container not in components:
                                components.append(container)

        return components


    # def map_lipid_maps_compounds_with_side_chain_rules(self,generic_compound,core,side_chain_atoms_rules={}, side_chain_molecule_rules={},n_side_chains = 0):
    #
    #     generic_compound_container = self.__lipid_maps_database.get_compound_by_id(generic_compound)
    #
    #     name = generic_compound_container.getName()
    #     formula = generic_compound_container.getFormula()
    #     db_id = generic_compound_container.getDbId()
    #     smiles = generic_compound_container.getSmiles()
    #     aliases = generic_compound_container.getAliases()
    #
    #     generic_compound_ont_id = self.__compounds_ontology.add_generic_node(name,formula,db_id,smiles,aliases = aliases)
    #
    #     smiles_database = self.__lipid_maps_database.getSmilesDatabase()
    #     smarts = MolFromSmarts(core)
    #     smiles_keys = list(smiles_database.keys())
    #
    #     filtered_compounds, _ = self.__get_filtered_compounds(smiles_keys,smarts,smiles_database)
    #
    #     for complete_compound in filtered_compounds:
    #         compound_mol = MolFromSmiles(complete_compound.getSmiles())
    #         sidechain_compound = AllChem.DeleteSubstructs(compound_mol, smarts)
    #         finalCheck = self.__process_and_check_side_chain(sidechain_compound, compound_mol,core,
    #                                                          side_chain_atoms_rules,side_chain_molecule_rules)
    #
    #         if finalCheck:
    #             sidechain, sidechains_smiles_list = ChemoUtilities.retrieve_fragments(sidechain_compound)
    #
    #             side_chain_counter = len(sidechain)
    #
    #             if n_side_chains==0 or side_chain_counter == n_side_chains:
    #
    #                 inchikey = complete_compound.getInchiKey()
    #                 modelseed_compound = self.__modelSeedCompoundsDatabase.get_compound_by_inchi_key(inchikey)
    #
    #                 id, name, formula, inchikey, smiles, aliases = \
    #                     self.__generate_complete_compound(complete_compound,modelseed_compound,compound_mol)
    #
    #                 complete_node_ont_id = \
    #                     self.__compounds_ontology.add_complete_node(id, name, inchikey, formula, smiles, aliases = aliases)
    #
    #                 self.__compounds_ontology.add_relationship(complete_node_ont_id,generic_compound_ont_id,"is_a")
    #
    #
    # def __process_and_check_side_chain(self, sidechains,compound_mol,core_smart,
    #                                    side_chain_atoms_rules,
    #                                    side_chain_molecule_rules):
    #     sidechains.UpdatePropertyCache()
    #     Chem.GetSymmSSSR(sidechains)
    #
    #     finalChecker = SidechainChecker()
    #     finalCheck = finalChecker(core_smart, compound_mol,
    #                               side_chain_atoms_rules,
    #                               side_chain_molecule_rules)
    #
    #     return finalCheck

    def __check_if_product_is_compatible_with_ont_container(self,model_seed_compounds,ont_ids):

        # model_seed_compounds = self.__convert_metacyc_compounds_into_model_seed(metacyc_compounds)

        res = []

        for ont_id in ont_ids:

            for compound in model_seed_compounds:

                parent = self.accessor.get_successors_by_ont_id_rel_type(ont_id,"is_a")
                if parent:

                    ms_id = compound.getDbId()
                    container = self.accessor.get_node_by_ont_id(parent[0])
                    aliases = container.aliases
                    if ModelSEEDDatabases.MODEL_SEED.value in aliases.keys():
                        for alias in aliases[ModelSEEDDatabases.MODEL_SEED.value]:
                            if ms_id == alias:
                                res.append(ont_id)


        return res

    def __simplify_compounds(self, modelseed_compounds):
        res = []
        for compound in modelseed_compounds:
            if compound in self.compounds_simplifier.keys():
                res.append(self.compounds_simplifier.get(compound))
            else:
                res.append(compound)

        return res

if __name__ == "__main__":

    coiso = OntologyGeneratorDA()
    core_fatty_acids = \
        "C(=O)O"
    core_alcohol = "*CO"

    reaction_transformation = "[*:1]C(=O)OC[C@@H](O)COP(=O)(O)OC[C@@H](O)CO.[*:2]C(=O)OC[C@H](COP(=O)(O)OC[C@@H](O)CO)OC([*:3])=O>>[*:1]C(=O)OC[C@@H](COP(O)(=O)OC[C@@H](COC([*:2])=O)OC([*:3])=O)O"
    coiso.create_new_compounds_using_transformation(reaction_transformation,[110,41],[765625],"LPLIPA6","BiGG",True,True,reverse=False)

    # coiso("TRIGLSYN-PWY", [core_fatty_acids,core_alcohol] , "transformations_TRIGLSYN-PWY", ["PWY-7277","PWY-7835"],["RXN-12383","RXN-1641"])

    # coiso.handle_reaction("RXN-15211","PWY-7277","transformations_Sphingolipid-Biosynthesis")
    # coiso.handle_reaction("RXN-15212", "PWY-7277", "transformations_Sphingolipid-Biosynthesis")

    # transformations_cardiolipin = TransformationsHandler()
    # transformations_cardiolipin.choose_transformations_from_pathway("TRIGLSYN-PWY", ["cpd27031", "cpd27029", "cpd00487"])
    # transformations_cardiolipin.save_transformations("transformations_TRIGLSYN-PWY")



import shutil
import urllib
from datetime import datetime

from boimmgpy.utilities import file_utilities
from boimmgpy.definitions import ROOT_DIR, DATABASE_CONFIGS


class ReactionsIDConverterScraper:

    def __init__(self):
        self.url_db = "https://raw.githubusercontent.com/ModelSEED/ModelSEEDDatabase/master/Biochemistry/Aliases/Unique_ModelSEED_Reaction_Aliases.txt"

        self.__database_config = file_utilities.read_conf_file(DATABASE_CONFIGS)
        up_to_date = self.__check_time()
        if not up_to_date:
            self.__download_modelseed_database()

    def __check_time(self):
        last_update_time = self.__database_config["time_model_seed_reactions_aliases"]
        last_update_time = datetime.strptime(last_update_time, "%Y-%m-%d")
        time_now = datetime.now()
        delta = time_now - last_update_time
        if delta.days >= 7:
            return False
        else:
            return True

    def __download_modelseed_database(self):

        with urllib.request.urlopen(self.url_db) as response, open(ROOT_DIR + self.__database_config["path_model_seed_reactions_aliases"],
                                                                'wb') as out_file:
            shutil.copyfileobj(response, out_file)

        self.__database_config["time_model_seed_reactions_aliases"] = str(datetime.now()).split(" ")[0]
        file_utilities.change_conf_file(DATABASE_CONFIGS, self.__database_config)

    def __call__(self):
        return ROOT_DIR + self.__database_config["path_model_seed_reactions_aliases"]


import re

from boimmgpy.kegg.kegg_entity import KeggEntity
import Bio.KEGG.REST as kegg_api


class KeggPathway(KeggEntity):

    def __init__(self,entry, byModules=True):
        super().__init__(entry)
        self.byModules = byModules
        self.__get_all_information()


    def __get_all_information(self):
        if self.byModules:
            info = self.get_raw_data().split("\n")
            self.modules = []
            go = False
            for text in info:

                if re.search("\AMODULE", text):
                    new_line = re.sub('\s+', "\t", text)
                    module = new_line.split("\t")[1]
                    self.modules.append(module)
                    go = True
                elif (re.search("\ADISEASE", text) or re.search("\ADBLINKS", text)
                    or re.search("\AREFERENCE", text) or re.search("\AKO_PATHWAY", text)):
                    go = False
                elif go:
                    new_line = re.sub('\s+', "\t", text)
                    module = new_line.split("\t")[1]
                    self.modules.append(module)
        else:

            data = kegg_api.kegg_link("reaction", "path:" + self.id)
            self.reactions = []
            for line in data:
                reaction = line.strip().split("\t")[1].split(":")[1]
                self.reactions.append(reaction)


    def get_modules(self):
        return self.modules

    def get_compounds(self):
        return self.compounds

    def get_reactions(self):
        return self.reactions


from cobra import Metabolite, Reaction, Model
from cobrababel import bigg

from boimmgpy.database.containers.compound_node import CompoundNode
from boimmgpy.id_converters.compounds_id_converter import CompoundsIDConverter
from boimmgpy.utilities.annotation_utils import AnnotationUtils
from boimmgpy.kegg.kegg_compound import KeggCompound
from boimmgpy.kegg.kegg_reaction import KeggReaction

import numpy as np


def check_if_elem_exist_in_list_of_cobra_containers(lst,elem):
    """
    Function to check if a given element is in a list with cobrapy containers (Metabolite, Reaction, etc)
    :param list lst: list of cobrapy containers
    :param elem: element
    :return boolean: if it exists in the list returns True, otherwise returns False
    """

    for el in lst:
       if elem.id == el.id:
           return True

    return False

def set_objective_function(model, reaction_id):

    reaction = model.reactions.get_by_id(reaction_id)
    model.objective = reaction

def evalSlimSol(solution, tol):

    if np.isnan(solution): return False

    if abs(solution) < tol: return False

    return True


def get_unique_list_of_cobra_containers(lst):
    """
    This function checks whether there are repetitive cobrapy containers in a given list and returns a list with unique
    containers.

    :param list lst: list of cobrapy containers
    :return list: list with unique cobrapy containers
    """

    res= []
    for elem in lst:
        if not check_if_elem_exist_in_list_of_cobra_containers(res,elem):
            res.append(elem)
    return res

#################################################### Generate model compounds ###############################################



def generate_model_compounds_by_database_format(model, modelseedid,
                                                compoundsIdConverter,
                                                modelseedCompoundsDb,
                                                database_format,
                                                compartment=""):

    """
    This function allows the generation of a given compound in the chosen database format.

    :param cobrapy.Model model: a cobrapy model
    :param string modelseedid: the model seed identifier
    :param CompoundsIDConverter compoundsIdConverter: a converter
    :param ModelSeedCompoundsDB modelseedCompoundsDb: a compound database
    :param string database_format: database name (only "ModelSEED", "BiGG" and "KEGG" supported)
    :param (Optional) string compartment: specific compartment to introduce compound
    :return list: a list of the generated compounds (in all the model compartments)
    """

    translation = compoundsIdConverter.get_modelSeedIdToDb().get(modelseedid)
    new_compounds = []
    if translation:
        if database_format == "BiGG" and \
                ("BiGG" in translation.keys() or "BiGG1" in translation.keys()):
            new_compounds = generate_bigg_compound(model,
                                                   modelseedid,
                                                   compoundsIdConverter,
                                                   modelseedCompoundsDb,
                                                   compartment)
            if new_compounds:
                return new_compounds

        elif database_format == "KEGG" and \
                "KEGG" in translation.keys():
            new_compounds = generate_kegg_compound(model,
                                                   modelseedid,
                                                   compoundsIdConverter,
                                                   modelseedCompoundsDb,
                                                   compartment)

            if new_compounds:
                return new_compounds

        if not new_compounds:
            new_compounds = generate_modelseed_compound(model,
                                                        modelseedid,
                                                        compoundsIdConverter,
                                                        modelseedCompoundsDb,
                                                        compartment)
            if new_compounds:
                return new_compounds

    else:
        new_compounds = generate_modelseed_compound(model,
                                                    modelseedid,
                                                    compoundsIdConverter,
                                                    modelseedCompoundsDb,
                                                    compartment)

        if new_compounds:
            return new_compounds


def generate_bigg_compound(model,model_seed_id,compoundsIdConverter,modelseedCompoundsDb,compartment=""):
    """
    Method to generate metabolites for each compartment of the model in the BiGG database format.
    This method tries to find the BiGG format of the new compounds. If it does not find, it will change it into the
    Model SEED format.

    :param cobrapy.Model model: cobrapy Model
    :param string model_seed_id: model seed identifier of the compound
    :param CompoundsIDConverter compoundsIdConverter: a converter
    :param ModelSeedCompoundsDB modelseedCompoundsDb: the compounds Model SEED database
    :param (Optional) string compartment: compartment to introduce the new compound
    :return list: new compounds list
    """

    if model_seed_id and "BiGG" in compoundsIdConverter.get_modelSeedIdToDb().get(model_seed_id)\
            or "BiGG1" in compoundsIdConverter.get_modelSeedIdToDb().get(model_seed_id):
        if "BiGG" in compoundsIdConverter.get_modelSeedIdToDb().get(model_seed_id):
            bigg_ids = compoundsIdConverter.convert_modelSeedId_into_other_dbID(model_seed_id, "BiGG")
        else:
            bigg_ids = compoundsIdConverter.convert_modelSeedId_into_other_dbID(model_seed_id, "BiGG1")

        bigg_metabolite  = None
        bigg_id = None
        found = False
        i = 0
        while not found and i < len(bigg_ids):
            try:
                bigg_id = bigg_ids[i]
                bigg_metabolite = bigg.get_bigg_metabolite(bigg_id)
                found = True
            except:
                i += 1

        if bigg_metabolite:
            compounds = []
            if not compartment:
                for compartment in model.compartments:
                    model_metabolite = Metabolite(bigg_id + "_" + compartment)
                    aliases = compoundsIdConverter.get_all_aliases_by_modelSeedID(model_seed_id)
                    annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(aliases)
                    model_metabolite.annotation = annotation
                    if bigg_metabolite.get("inchikey"):
                        bigg_inchikey = bigg_metabolite.get("inchikey")
                        if bigg_inchikey:
                            model_metabolite.annotation["inchikey"] = bigg_inchikey

                    model_metabolite.formula = bigg_metabolite.get("formulae")[0]
                    model_metabolite.charge =  bigg_metabolite.get("charges")[0]
                    model_metabolite.name = bigg_metabolite.get("name")
                    model_metabolite.compartment = compartment
                    compounds.append(model_metabolite)

            else:
                model_metabolite = Metabolite(bigg_id + "_" + compartment)
                aliases = compoundsIdConverter.get_all_aliases_by_modelSeedID(model_seed_id)
                annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(aliases)
                model_metabolite.annotation = annotation
                if bigg_metabolite.get("inchikey"):
                    model_metabolite.annotation["inchikey"] = bigg_metabolite.get("inchikey")

                model_metabolite.formula = bigg_metabolite.get("formulae")[0]
                model_metabolite.charge = bigg_metabolite.get("charges")[0]
                model_metabolite.name = bigg_metabolite.get("name")
                model_metabolite.compartment = compartment
                compounds.append(model_metabolite)


            return compounds

        else:
            return generate_modelseed_compound(model,model_seed_id,compoundsIdConverter,modelseedCompoundsDb)

    else:
        return generate_modelseed_compound(model,model_seed_id,compoundsIdConverter,modelseedCompoundsDb)



def generate_modelseed_compound(model,model_seed_id,compoundsIdConverter,modelseedCompoundsDb, compartment=""):
    """
    Method to generate metabolites for each compartment of the model in the Model SEED database format.
    This method tries to find the Model SEED format of the new compounds.

    :param cobrapy.Model model: cobrapy Model
    :param string model_seed_id: model seed identifier of the compound
    :param CompoundsIDConverter compoundsIdConverter: a converter
    :param ModelSeedCompoundsDB modelseedCompoundsDb: the compounds Model SEED database
    :param (Optional) string compartment: compartment to introduce the new compound
    :return list: new compounds list
    """

    modelseed_compound = modelseedCompoundsDb.get_compound_by_id(model_seed_id)

    aliases = compoundsIdConverter.get_all_aliases_by_modelSeedID(model_seed_id)
    annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(aliases)
    compounds = []

    if not compartment:

        for compartment in model.compartments:
            db_id = modelseed_compound.getDbId()
            model_metabolite = Metabolite(db_id + "_" + compartment)
            model_metabolite.annotation = annotation
            model_metabolite.annotation["inchikey"] = modelseed_compound.getInchikey()
            model_metabolite.formula = modelseed_compound.getFormula()
            model_metabolite.charge = modelseed_compound.getCharge()
            model_metabolite.name = modelseed_compound.getName()
            model_metabolite.compartment=compartment
            compounds.append(model_metabolite)

    else:
        db_id = modelseed_compound.getDbId()
        model_metabolite = Metabolite(db_id + "_" + compartment)
        model_metabolite.annotation = annotation
        model_metabolite.annotation["inchikey"] = modelseed_compound.getInchikey()
        model_metabolite.formula = modelseed_compound.getFormula()
        model_metabolite.charge = modelseed_compound.getCharge()
        model_metabolite.name = modelseed_compound.getName()
        model_metabolite.compartment = compartment
        compounds.append(model_metabolite)

    return compounds

def generate_boimmg_metabolites(model : Model, compound : CompoundNode,
                                database_format: str, compoundsIdConverter : CompoundsIDConverter,
                                compoundsAnnotationConfigs) -> list:
    """
    Generate an arbitrary format for the metabolites

    :param model:
    :param compound:
    :param database_format:
    :param compoundsIdConverter:
    :return:
    """

    compounds = []
    modelseedid = compound.model_seed_id

    id = compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]+ str(compound.id)

    formula = compound.formula
    name = compound.name
    charge = compound.charge
    if charge:
        charge = int(charge)

    model_compartments = model.compartments

    for compartment in model_compartments:
        new_metabolite = Metabolite(id=id+"_"+compartment, name=name, charge=charge, formula=formula)
        annotation = {}
        if modelseedid:
            aliases = compoundsIdConverter.get_all_aliases_by_modelSeedID(modelseedid)

            annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(aliases)

        annotation["boimmg.compound"] = compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"] + \
                                        str(compound.id)

        new_metabolite.annotation = annotation
        new_metabolite.annotation["smiles"] = compound.smiles
        new_metabolite.annotation["inchikey"] = compound.inchikey
        new_metabolite.compartment = compartment
        compounds.append(new_metabolite)

    return compounds

def generate_boimmg_metabolites_in_compartment(compound : CompoundNode,
                                compartment: str, compoundsIdConverter : CompoundsIDConverter,
                                compoundsAnnotationConfigs) -> Metabolite:
    """
    Generate an arbitrary format for the metabolites

    :param model:
    :param compound:
    :param database_format:
    :param compoundsIdConverter:
    :return:
    """

    modelseedid = compound.model_seed_id

    id = compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]+ str(compound.id)

    formula = compound.formula
    name = compound.name
    charge = compound.charge

    if charge:
        charge = int(charge)
    else:
        charge = 0

    new_metabolite = Metabolite(id=id+"_"+compartment, name=name, charge=charge, formula=formula)

    aliases = compound.aliases
    annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(aliases)
    if modelseedid:
        aliases = compoundsIdConverter.get_all_aliases_by_modelSeedID(modelseedid)

        annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(aliases)

    annotation["boimmg.compound"] = compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"] + \
                                    str(compound.id)

    new_metabolite.annotation = annotation
    new_metabolite.annotation["smiles"] = compound.smiles
    new_metabolite.annotation["inchikey"] = compound.inchikey
    new_metabolite.compartment = compartment

    return new_metabolite


def generate_kegg_compound(model,modelseedid,compoundsIdConverter,modelseedCompoundsDb, compartment=""):
    """
    Method to generate metabolites for each compartment of the model in the KEGG database format.
    This method tries to find the KEGG format of the new compounds. If it does not find, it will change it into the
    Model SEED format.

    :param cobrapy.Model model: cobrapy Model
    :param string modelseedid: model seed identifier of the compound
    :param CompoundsIDConverter compoundsIdConverter: a converter
    :param ModelSeedCompoundsDB modelseedCompoundsDb: the compounds Model SEED database
    :param (Optional) string compartment: compartment to introduce the new compound
    :return list: new compounds list
    """

    if modelseedid and "KEGG" in compoundsIdConverter.get_modelSeedIdToDb().get(modelseedid):
        kegg_id = compoundsIdConverter.convert_modelSeedId_into_other_dbID(modelseedid, "KEGG")[0]
        kegg_metabolite = KeggCompound(kegg_id)

        aliases = compoundsIdConverter.get_all_aliases_by_modelSeedID(modelseedid)
        annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(aliases)

        compounds = []
        if not compartment:
            for compartment in model.compartments:

                model_metabolite = Metabolite(kegg_id + "_" + compartment)
                model_metabolite.annotation = annotation
                model_metabolite.annotation["inchikey"] = kegg_metabolite.get_inchikey()
                model_metabolite.annotation["smiles"] = kegg_metabolite.get_smiles()
                model_metabolite.formula = kegg_metabolite.get_formula()
                model_metabolite.name = kegg_metabolite.get_name()
                model_metabolite.charge = 0
                model_metabolite.compartment = compartment
                compounds.append(model_metabolite)

        else:
            model_metabolite = Metabolite(kegg_id + "_" + compartment)
            model_metabolite.annotation = annotation
            model_metabolite.annotation["inchikey"] = kegg_metabolite.get_inchikey()
            model_metabolite.annotation["smiles"] = kegg_metabolite.get_smiles()
            model_metabolite.formula = kegg_metabolite.get_formula()
            model_metabolite.name = kegg_metabolite.get_name()
            model_metabolite.charge = 0
            model_metabolite.compartment = compartment
            compounds.append(model_metabolite)

        return compounds



    else:
        return generate_modelseed_compound(model, modelseedid, compoundsIdConverter,modelseedCompoundsDb)



#################################################### Generate model reactions ###############################################


def generate_reactions_to_model(model, reactions, babel, compartment,
                                compoundsIdConverter,modelseedCompoundsDb,
                                reactionsIdConverter,database_format, compoundsAnnotationConfigs):

    """
    This function will generate model reactions taking into account the :param reactions and the :param babel.
    The :param babel is a dictionary carrying the information about the compounds swapped in the model.
    It will generate the reactions that have the available reactants and products in a specific model compartment

    :param cobrapy.Model model: a cobrapy model
    :param list reactions: a list of reactions to possibly add to the model
    :param dict babel: a dictionary with the model seed id as a key and a list of model compounds.
    :param Model.Compartment compartment:
    :param CompoundsIDConverter compoundsIdConverter: a compounds ID converter
    :param ModelSeedCompoundsDB modelseedCompoundsDb: the Model SEED compounds database
    :param ReactionsIDConverter reactionsIdConverter: a reactions ID converter
    :param string database_format: database name (only "ModelSEED", "BiGG" and "KEGG" supported)
    :return:
    """

    reactions_to_add = []

    for modelseed_reaction in reactions:

        compounds = modelseed_reaction.getCompounds()
        reaction_found_compounds = babel.copy()
        for compound in compounds:
            aliases = compoundsIdConverter.get_all_aliases_by_modelSeedID(compound)
            modelseed_compound = modelseedCompoundsDb.get_compound_by_id(compound)
            inchikey = modelseed_compound.getInchikey()

            compounds_in_model = check_if_metabolite_exists_in_model(model,inchikey,aliases,compoundsAnnotationConfigs)
            if compounds_in_model:
                i=0
                found=False
                compartment_compound = None
                while not found and i<len(compounds_in_model):
                    if compounds_in_model[i].compartment == compartment:
                        found =True
                        compartment_compound = compounds_in_model[i]
                    i+=1

                if compartment_compound:
                    reaction_found_compounds[compound] = compounds_in_model

        if len(reaction_found_compounds.keys()) == len(compounds):
            print(modelseed_reaction.getName())
            translation = reactionsIdConverter.get_modelSeedIdToDb().get(modelseed_reaction.getDbId())
            new_reaction=None
            if translation:
                if database_format == "BiGG" and \
                        ("BiGG" in translation.keys() or "BiGG1" in translation.keys()):
                    new_reaction = generate_bigg_reaction(modelseed_reaction,reaction_found_compounds,
                                                          compartment,reactionsIdConverter,compoundsIdConverter)
                    if new_reaction:
                        reactions_to_add.append(new_reaction)

                elif database_format == "KEGG" and \
                        "KEGG" in translation.keys():
                    new_reaction = generate_kegg_reaction(modelseed_reaction,
                                                          reaction_found_compounds,
                                                          compartment,
                                                          reactionsIdConverter,compoundsIdConverter)
                    if new_reaction:
                        reactions_to_add.append(new_reaction)

                if not new_reaction:
                    new_reaction = generate_modelseed_reaction(modelseed_reaction,
                                                               reaction_found_compounds,
                                                               compartment,
                                                               reactionsIdConverter)
                    if new_reaction:
                        reactions_to_add.append(new_reaction)

            else:
                new_reaction = generate_modelseed_reaction(modelseed_reaction,
                                                           reaction_found_compounds,
                                                           compartment,
                                                           reactionsIdConverter)
                if new_reaction:
                    reactions_to_add.append(new_reaction)

    model.add_reactions(reactions_to_add)


def generate_modelseed_reaction(modelseed_reaction,reaction_found_compounds,compartment,reactionsIdConverter):
    """
    This function generates a new Model SEED format reaction in a given compartment.

    :param ModelSeedReaction modelseed_reaction: a Model SEED reaction container
    :param dict reaction_found_compounds: dictionary with model seed compounds identifiers as keys and the
    respective Model.Metabolite as value.
    :param Model.Compartment compartment:
    :param ReactionsIDConverter reactionsIdConverter: a reaction identifier converter
    :return cobrapy.Reaction: the new model reaction
    """

    stoichiometry = modelseed_reaction.getStoichiometry()
    new_stoichiometry = {}

    for compound_id in stoichiometry:
        if compound_id in reaction_found_compounds.keys():
            model_compound = reaction_found_compounds[compound_id]
            new_stoichiometry[model_compound] = stoichiometry[compound_id]

    id = modelseed_reaction.getDbId() + "_" + compartment
    name = modelseed_reaction.getName()
    reversibility = modelseed_reaction.getReversibility()
    direction = modelseed_reaction.getDirection()
    if direction == "<":
        for compound in new_stoichiometry:
            new_stoichiometry[compound] = new_stoichiometry[compound] * -1

    if reversibility:
        new_model_reaction = Reaction(id,name,lower_bound=None)

    else:
        new_model_reaction = Reaction(id,name)

    aliases = reactionsIdConverter.get_all_aliases_by_modelSeedID(modelseed_reaction.getDbId())

    if aliases:
        new_model_reaction.annotation = AnnotationUtils.get_reaction_annotation_format_by_aliases(aliases)
    new_model_reaction.add_metabolites(new_stoichiometry)

    return new_model_reaction


def generate_bigg_reaction(modelseed_reaction,reaction_found_compounds,
                           compartment,reactionsIdConverter,
                           compoundsIdConverter):
    """
    This function generates a new BiGG format reaction in a given compartment.

    :param ModelSeedReaction modelseed_reaction: a Model SEED reaction container
    :param dict reaction_found_compounds: dictionary with model seed compounds identifiers as keys and the
    respective Model.Metabolite as value.
    :param Model.Compartment compartment:
    :param ReactionsIDConverter reactionsIdConverter: a reaction identifier converter
    :param CompoundsIDConverter compoundsIdConverter: a compounds identifier converter
    :return cobrapy.Reaction: the new model reaction
    """

    bigg_reaction_ids = \
    reactionsIdConverter.convert_modelSeedId_into_other_dbID(modelseed_reaction.getDbId(), "BiGG")

    bigg_reaction = None
    bigg_id = None
    found = False
    i = 0
    while not found and i < len(bigg_reaction_ids):
        try:
            bigg_reaction = bigg.get_bigg_reaction(bigg_reaction_ids[i])
            bigg_id = bigg_reaction_ids[i]
            found = True
        except:
            i += 1

    if bigg_reaction:
        bigg_metabolites = bigg_reaction.get("metabolites")

        new_stoichiometry = {}

        for compound in bigg_metabolites:
            compound_id = compound.get("bigg_id")
            modelseedid = compoundsIdConverter.convert_dbID_into_modelSeedId("BiGG", compound_id)[0]
            model_compound = reaction_found_compounds[modelseedid]
            new_stoichiometry[model_compound] = compound.get("stoichiometry")

        id = bigg_id + "_" + compartment
        name = bigg_reaction.get("name")
        reversibility = modelseed_reaction.getReversibility()

        if reversibility:
            new_reaction_model = Reaction(id, name, lower_bound=None)

        else:
            new_reaction_model = Reaction(id, name)

        new_reaction_model.add_metabolites(new_stoichiometry)
        aliases = reactionsIdConverter.get_all_aliases_by_modelSeedID(modelseed_reaction.getDbId())
        new_reaction_model.annotation = AnnotationUtils.get_reaction_annotation_format_by_aliases(aliases)
        return new_reaction_model


def generate_kegg_reaction(modelseed_reaction,reaction_found_compounds,
                             compartment,reactionsIdConverter,compoundsIdConverter):
    """
    This function generates a new KEGG format reaction in a given compartment.

    :param ModelSeedReaction modelseed_reaction: a Model SEED reaction container
    :param dict reaction_found_compounds: dictionary with model seed compounds identifiers as keys and the
    respective Model.Metabolite as value.
    :param Model.Compartment compartment:
    :param ReactionsIDConverter reactionsIdConverter: a reaction identifier converter
    :param CompoundsIDConverter compoundsIdConverter: a compounds identifier converter
    :return cobrapy.Reaction: the new model reaction
    """

    kegg_reaction_id = reactionsIdConverter.convert_modelSeedId_into_other_dbID(modelseed_reaction.getDbId(),"KEGG")[0]
    try:
        kegg_reaction = KeggReaction(kegg_reaction_id)

        stoichiometry = kegg_reaction.get_stoichiometry()
        new_stoichiometry = {}

        for compound_id in stoichiometry:
            modelseedids = compoundsIdConverter.convert_dbID_into_modelSeedId("KEGG",compound_id)

            found = False
            i=0
            found_compound = None
            while not found and i<len(modelseedids):
                if modelseedids[i] in reaction_found_compounds.keys():
                    found = True
                    found_compound = modelseedids[i]

                i+=1

            if found:
                model_compound = reaction_found_compounds[found_compound]
                new_stoichiometry[model_compound] = stoichiometry[compound_id]

            else:
                return None

        id = kegg_reaction.get_id() + "_" + compartment

        name = kegg_reaction.get_name()
        if not name:
            name = modelseed_reaction.getName()

        reversibility = kegg_reaction.get_reversibility()
        direction = kegg_reaction.get_direction()
        if direction == "<=":
            for compound in new_stoichiometry:
                new_stoichiometry[compound] = new_stoichiometry[compound] * -1

        if reversibility:
            new_model_reaction = Reaction(id, name, lower_bound=None)

        else:
            new_model_reaction = Reaction(id, name)

        aliases = reactionsIdConverter.get_all_aliases_by_modelSeedID(modelseed_reaction.getDbId())
        new_model_reaction.annotation = AnnotationUtils.get_reaction_annotation_format_by_aliases(aliases)
        new_model_reaction.add_metabolites(new_stoichiometry)
        return new_model_reaction

    except:
        return None


def generate_reaction_in_compartimentalized_model(modelseed_reaction,
                                                  compartment, babel, reactionsIdConverter,
                                                  compoundsIdConverter, database_format):

    """
    This function will generate a model reaction taking into account the :param reactions and the :param babel.
    The :param babel is a dictionary carrying the information about the compounds swapped in the model.
    It will generate one reaction that have the available reactants and products in a specific model compartment.

    :param ModelSeedReaction modelseed_reaction:
    :param Model.Compartment compartment:
    :param dict babel: dictionary with specific information about some metabolites in the model
    :param CompoundsIDConverter compoundsIdConverter: a compounds ID converter
    :param ReactionsIDConverter reactionsIdConverter: a reactions ID converter
    :param string database_format: database name (only "ModelSEED", "BiGG" and "KEGG" supported)
    :return cobrapy.Reaction: new reaction

    """

    translation = reactionsIdConverter.get_modelSeedIdToDb().get(modelseed_reaction.getDbId())
    new_reaction = None
    if translation:
        if database_format == "BiGG" and \
                ("BiGG" in translation.keys() or "BiGG1" in translation.keys()):

            new_reaction = generate_bigg_reaction(modelseed_reaction,
                                                                 babel,
                                                                 compartment,
                                                                 reactionsIdConverter,
                                                                 compoundsIdConverter)
            if new_reaction:
                return new_reaction

        elif database_format == "KEGG" and \
                "KEGG" in translation.keys():
            new_reaction = generate_kegg_reaction(modelseed_reaction,babel,compartment,
                                                                 reactionsIdConverter,
                                                                 compoundsIdConverter)
            if new_reaction:
                return new_reaction

        if not new_reaction:
            new_reaction = generate_modelseed_reaction(modelseed_reaction,
                                                                      babel,compartment,
                                                                      reactionsIdConverter)
            if new_reaction:
                return new_reaction

    else:
        new_reaction = generate_modelseed_reaction(modelseed_reaction,
                                                                  babel,
                                                                  compartment,
                                                                  reactionsIdConverter)
        if new_reaction:
            return new_reaction



########################################################## Checkers ####################################################

def check_if_reaction_exists_in_model(model, modelseed_reaction,reactionsIdConverter,reactionsAnnotationConfigs):
    """
    This function checks whether a given reaction is present in a given model or not.
    It returns the model reaction or None if not found
    :param model:
    :param modelseed_reaction:
    :param reactionsIdConverter:
    :param reactionsAnnotationConfigs:
    :return:
    """

    aliases = reactionsIdConverter.get_all_aliases_by_modelSeedID(modelseed_reaction.getDbId())

    for reaction in model.reactions:
        annotation = reaction.annotation

        for database in aliases:

            if database in reactionsAnnotationConfigs.keys():
                annotation_db = reactionsAnnotationConfigs[database]
            else:
                annotation_db = database

            if annotation_db in annotation.keys():
                if type(annotation[annotation_db]) == list:
                    for alias in annotation[annotation_db]:
                        if alias in aliases[database]:
                            return reaction
                elif annotation[annotation_db] in aliases[database]:
                    return reaction
    return None


def check_if_metabolite_exists_in_model(model, inchikey, aliases, compoundsAnnotationConfigs, boimmg_container =None):
    """
    This method searches for a given metabolite in the Model

    :param string inchikey: InChIKey of the metabolite to be found in the model
    :param dict aliases: dictionary with the metabolite aliases ({database : list with aliases})

    :return Model.metabolite: if the metabolite is found returns the metabolite, otherwise returns None
    """


    metabolites = []
    for metabolite in model.metabolites:
        found = False

        if inchikey:
            if "inchikey" in metabolite.annotation.keys():
                inchi_key = metabolite.annotation.get("inchikey")
                if inchi_key and inchi_key!="":
                    if inchi_key[:-1] == inchikey[:-1]:
                        # found_inchi_key_metabolite = True
                        if metabolite not in metabolites:
                            metabolites.append(metabolite)

        if not found:
            for alias in aliases:

                if alias in compoundsAnnotationConfigs.keys():
                    new_alias = compoundsAnnotationConfigs[alias]

                    if new_alias in metabolite.annotation.keys():
                        aliases_in_model = metabolite.annotation.get(new_alias)

                        if type(aliases_in_model) == list:
                            for alias_in_model in aliases_in_model:
                                if alias_in_model in aliases[alias]:
                                    # found_alias = True
                                    if metabolite not in metabolites:
                                        metabolites.append(metabolite)


                        elif aliases_in_model in aliases[alias]:
                            if metabolite not in metabolites:
                                metabolites.append(metabolite)

        if not found and boimmg_container:
            new_alias = compoundsAnnotationConfigs["BOIMMG"]

            if new_alias in metabolite.annotation.keys():
                aliases_in_model = metabolite.annotation.get(new_alias)
                boimmg_id = compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]+str(boimmg_container.id)
                if aliases_in_model == boimmg_id:
                    metabolites.append(metabolite)


    return metabolites

############################################################################ Other ################################################################################

def get_compartments_babel(babel):
    """
    This function trawl through the babel in order to organize them by compartments.

    :param dict babel: a dictionary with the model seed id as a key and a list of model compounds.
    :return dict: dictionary of a dictionary with the compartments as keys and the babel for the compounds
    """
    res = {}
    modelseedids = list(babel.keys())
    for compound_in_model in babel[modelseedids[0]]:
        compartment = compound_in_model.compartment

        compartment_dict = {}
        compartment_dict[modelseedids[0]]=compound_in_model
        for i in range(1,len(modelseedids)):

            found = False
            j = 0
            compounds_in_model2 = babel[modelseedids[i]]
            while not found and j <len(compounds_in_model2):
                compound_compartment = compounds_in_model2[j].compartment
                if compound_compartment is None:
                    compound_compartment = "c"
                if compound_compartment == compartment:
                    compartment_dict[modelseedids[i]] = compounds_in_model2[j]
                    found=True
                j+=1

        if len(compartment_dict.keys()) == len(modelseedids):
            res[compartment] = compartment_dict
    return res

def check_if_is_exchange_reaction(reaction):
    """
    This function checks whether a given reaction is an exchange reaction.

    :param reaction:
    :return boolean:
    """
    reactants=False
    products = False
    stoich = reaction.metabolites
    for metabolite in stoich:
        if stoich[metabolite] < 0:
            reactants = True
        elif stoich[metabolite] > 0:
            products = True
    if reactants and products:
        return False
    else:
        return True

def swap_only_metabolites_in_reaction(metabolite, other, reaction):
    """
    This method swap a given metabolite into another in a specific reaction.

    :param Model.Metabolite metabolite: metabolite to be changed
    :param Model.Metabolite other: metabolite that will replace the other
    :param Model.Reaction reaction: reaction where the metabolites will be swapped.
    :return:
    """

    met_to_add = {}
    met_to_subtract = {}
    stoichiometry = reaction.metabolites
    coef = stoichiometry.get(metabolite)
    met_to_add[other] = coef
    met_to_subtract[metabolite] = coef
    reaction.subtract_metabolites(met_to_subtract)
    reaction.add_metabolites(met_to_add)

def get_model_metabolites_by_name(model,names):
    res=[]
    for name in names:
        res.extend(model.metabolites.query(name, attribute='name'))
    return res

def get_model_seed_id_by_model_compound(compound,configs,converter) -> list:

    annotation = compound.annotation
    model_seed_annotation_key = configs.get("ModelSEED")
    if model_seed_annotation_key in annotation.keys():

        model_seed_id = annotation.get(model_seed_annotation_key)

        if type(model_seed_id) == str:
            return model_seed_id

        else:
            return model_seed_id[0]

    else:
        annotation_keys = list(configs.keys())
        for key in annotation_keys:
            annotation_value = configs.get(key)
            if annotation_value in annotation.keys():
                alias = annotation.get(annotation_value)

                if type(alias) == str:
                    model_seed_id =  converter.convert_db_id_to_model_seed_by_db_id(alias)
                    if model_seed_id:
                        return model_seed_id

                else:
                    i=0
                    while i<len(alias):
                        db_id = alias[i]
                        model_seed_id = converter.convert_db_id_to_model_seed_by_db_id(db_id)
                        if model_seed_id:
                            return model_seed_id
                        i+=1
    return None

def convert_model_seed_reactions_into_model_reactions(model,reactions,
                                                      reactionsIdConverter,reactionsAnnotationConfigs,
                                                      compoundsIdConverter,modelseedCompoundsDb,
                                                      compoundsAnnotationConfigs,database_format,
                                                      model_container=None,modelseedid = None):


    filtered_reactions = []
    for reaction in reactions:
        inModel = check_if_reaction_exists_in_model(model,
                                                    reaction,
                                                    reactionsIdConverter,
                                                    reactionsAnnotationConfigs)

        go = False
        if not inModel:
            stoichiometry = reaction.getStoichiometry()

            if model_container and modelseedid:

                if stoichiometry[modelseedid] > 0:
                    go = True

            else:
                go = True

            if go:
                compounds = reaction.getCompounds()
                reaction_found_compounds = {}

                for compound in compounds:
                    aliases = compoundsIdConverter.get_all_aliases_by_modelSeedID(compound)
                    modelseed_compound = modelseedCompoundsDb.get_compound_by_id(compound)
                    inchikey = modelseed_compound.getInchikey()

                    compound_in_model = check_if_metabolite_exists_in_model(model,
                                                                            inchikey,
                                                                            aliases,
                                                                            compoundsAnnotationConfigs)
                    if compound_in_model:
                        reaction_found_compounds[compound] = compound_in_model

                if len(reaction_found_compounds.keys()) == len(compounds):
                    print(reaction.getName(), reaction.getDbId())
                    compartment_babel = get_compartments_babel(reaction_found_compounds)
                    for compartment in compartment_babel:
                        new_reaction = generate_reaction_in_compartimentalized_model(reaction,
                                                                                compartment,
                                                                                compartment_babel[
                                                                                compartment],
                                                                                reactionsIdConverter,
                                                                                compoundsIdConverter,
                                                                                database_format)

                        if new_reaction and not check_if_is_exchange_reaction(new_reaction):

                            # try:
                            #     model.reactions.get_by_id(new_reaction.id)
                            # except:


                            filtered_reactions.append(new_reaction)

    return filtered_reactions






import os

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
TOOL_CONFIG_PATH = os.path.join(ROOT_DIR, 'configs/tool_configs.conf')
COMPOUNDS_ANNOTATION_CONFIGS_PATH = os.path.join(ROOT_DIR,"configs/compounds_annotation_configs.conf")
REACTIONS_ANNOTATION_CONFIGS_PATH = os.path.join(ROOT_DIR,"configs/reactions_annotation_configs.conf")
DATABASE_CONFIGS = os.path.join(ROOT_DIR,"configs/databases_files_settings.conf")
EXCEPTIONS = os.path.join(ROOT_DIR,"configs/generic_compounds_exceptions.conf")
BOIMMG_DATABASE = os.path.join(ROOT_DIR,"configs/database_settings.conf")